"""State module for managing EC2 Launch Template."""
import copy
import datetime
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    launch_template_data: make_dataclass(
        "RequestLaunchTemplateData",
        [
            ("KernelId", str, field(default=None)),
            ("EbsOptimized", bool, field(default=None)),
            (
                "IamInstanceProfile",
                make_dataclass(
                    "LaunchTemplateIamInstanceProfileSpecificationRequest",
                    [
                        ("Arn", str, field(default=None)),
                        ("Name", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "BlockDeviceMappings",
                List[
                    make_dataclass(
                        "LaunchTemplateBlockDeviceMappingRequest",
                        [
                            ("DeviceName", str, field(default=None)),
                            ("VirtualName", str, field(default=None)),
                            (
                                "Ebs",
                                make_dataclass(
                                    "LaunchTemplateEbsBlockDeviceRequest",
                                    [
                                        ("Encrypted", bool, field(default=None)),
                                        (
                                            "DeleteOnTermination",
                                            bool,
                                            field(default=None),
                                        ),
                                        ("Iops", int, field(default=None)),
                                        ("KmsKeyId", str, field(default=None)),
                                        ("SnapshotId", str, field(default=None)),
                                        ("VolumeSize", int, field(default=None)),
                                        ("VolumeType", str, field(default=None)),
                                        ("Throughput", int, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            ("NoDevice", str, field(default=None)),
                        ],
                    )
                ],
                field(default=None),
            ),
            (
                "NetworkInterfaces",
                List[
                    make_dataclass(
                        "LaunchTemplateInstanceNetworkInterfaceSpecificationRequest",
                        [
                            ("AssociateCarrierIpAddress", bool, field(default=None)),
                            ("AssociatePublicIpAddress", bool, field(default=None)),
                            ("DeleteOnTermination", bool, field(default=None)),
                            ("Description", str, field(default=None)),
                            ("DeviceIndex", int, field(default=None)),
                            ("Groups", List[str], field(default=None)),
                            ("InterfaceType", str, field(default=None)),
                            ("Ipv6AddressCount", int, field(default=None)),
                            (
                                "Ipv6Addresses",
                                List[
                                    make_dataclass(
                                        "InstanceIpv6AddressRequest",
                                        [("Ipv6Address", str, field(default=None))],
                                    )
                                ],
                                field(default=None),
                            ),
                            ("NetworkInterfaceId", str, field(default=None)),
                            ("PrivateIpAddress", str, field(default=None)),
                            (
                                "PrivateIpAddresses",
                                List[
                                    make_dataclass(
                                        "PrivateIpAddressSpecification",
                                        [
                                            ("Primary", bool, field(default=None)),
                                            (
                                                "PrivateIpAddress",
                                                str,
                                                field(default=None),
                                            ),
                                        ],
                                    )
                                ],
                                field(default=None),
                            ),
                            (
                                "SecondaryPrivateIpAddressCount",
                                int,
                                field(default=None),
                            ),
                            ("SubnetId", str, field(default=None)),
                            ("NetworkCardIndex", int, field(default=None)),
                            (
                                "Ipv4Prefixes",
                                List[
                                    make_dataclass(
                                        "Ipv4PrefixSpecificationRequest",
                                        [("Ipv4Prefix", str, field(default=None))],
                                    )
                                ],
                                field(default=None),
                            ),
                            ("Ipv4PrefixCount", int, field(default=None)),
                            (
                                "Ipv6Prefixes",
                                List[
                                    make_dataclass(
                                        "Ipv6PrefixSpecificationRequest",
                                        [("Ipv6Prefix", str, field(default=None))],
                                    )
                                ],
                                field(default=None),
                            ),
                            ("Ipv6PrefixCount", int, field(default=None)),
                        ],
                    )
                ],
                field(default=None),
            ),
            ("ImageId", str, field(default=None)),
            ("InstanceType", str, field(default=None)),
            ("KeyName", str, field(default=None)),
            (
                "Monitoring",
                make_dataclass(
                    "LaunchTemplatesMonitoringRequest",
                    [("Enabled", bool, field(default=None))],
                ),
                field(default=None),
            ),
            (
                "Placement",
                make_dataclass(
                    "LaunchTemplatePlacementRequest",
                    [
                        ("AvailabilityZone", str, field(default=None)),
                        ("Affinity", str, field(default=None)),
                        ("GroupName", str, field(default=None)),
                        ("HostId", str, field(default=None)),
                        ("Tenancy", str, field(default=None)),
                        ("SpreadDomain", str, field(default=None)),
                        ("HostResourceGroupArn", str, field(default=None)),
                        ("PartitionNumber", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            ("RamDiskId", str, field(default=None)),
            ("DisableApiTermination", bool, field(default=None)),
            ("InstanceInitiatedShutdownBehavior", str, field(default=None)),
            ("UserData", str, field(default=None)),
            (
                "TagSpecifications",
                List[
                    make_dataclass(
                        "LaunchTemplateTagSpecificationRequest",
                        [
                            ("ResourceType", str, field(default=None)),
                            (
                                "Tags",
                                List[
                                    make_dataclass(
                                        "Tag",
                                        [
                                            ("Key", str, field(default=None)),
                                            ("Value", str, field(default=None)),
                                        ],
                                    )
                                ],
                                field(default=None),
                            ),
                        ],
                    )
                ],
                field(default=None),
            ),
            (
                "ElasticGpuSpecifications",
                List[make_dataclass("ElasticGpuSpecification", [("Type", str)])],
                field(default=None),
            ),
            (
                "ElasticInferenceAccelerators",
                List[
                    make_dataclass(
                        "LaunchTemplateElasticInferenceAccelerator",
                        [("Type", str), ("Count", int, field(default=None))],
                    )
                ],
                field(default=None),
            ),
            ("SecurityGroupIds", List[str], field(default=None)),
            ("SecurityGroups", List[str], field(default=None)),
            (
                "InstanceMarketOptions",
                make_dataclass(
                    "LaunchTemplateInstanceMarketOptionsRequest",
                    [
                        ("MarketType", str, field(default=None)),
                        (
                            "SpotOptions",
                            make_dataclass(
                                "LaunchTemplateSpotMarketOptionsRequest",
                                [
                                    ("MaxPrice", str, field(default=None)),
                                    ("SpotInstanceType", str, field(default=None)),
                                    ("BlockDurationMinutes", int, field(default=None)),
                                    ("ValidUntil", datetime, field(default=None)),
                                    (
                                        "InstanceInterruptionBehavior",
                                        str,
                                        field(default=None),
                                    ),
                                ],
                            ),
                            field(default=None),
                        ),
                    ],
                ),
                field(default=None),
            ),
            (
                "CreditSpecification",
                make_dataclass("CreditSpecificationRequest", [("CpuCredits", str)]),
                field(default=None),
            ),
            (
                "CpuOptions",
                make_dataclass(
                    "LaunchTemplateCpuOptionsRequest",
                    [
                        ("CoreCount", int, field(default=None)),
                        ("ThreadsPerCore", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "CapacityReservationSpecification",
                make_dataclass(
                    "LaunchTemplateCapacityReservationSpecificationRequest",
                    [
                        ("CapacityReservationPreference", str, field(default=None)),
                        (
                            "CapacityReservationTarget",
                            make_dataclass(
                                "CapacityReservationTarget",
                                [
                                    ("CapacityReservationId", str, field(default=None)),
                                    (
                                        "CapacityReservationResourceGroupArn",
                                        str,
                                        field(default=None),
                                    ),
                                ],
                            ),
                            field(default=None),
                        ),
                    ],
                ),
                field(default=None),
            ),
            (
                "LicenseSpecifications",
                List[
                    make_dataclass(
                        "LaunchTemplateLicenseConfigurationRequest",
                        [("LicenseConfigurationArn", str, field(default=None))],
                    )
                ],
                field(default=None),
            ),
            (
                "HibernationOptions",
                make_dataclass(
                    "LaunchTemplateHibernationOptionsRequest",
                    [("Configured", bool, field(default=None))],
                ),
                field(default=None),
            ),
            (
                "MetadataOptions",
                make_dataclass(
                    "LaunchTemplateInstanceMetadataOptionsRequest",
                    [
                        ("HttpTokens", str, field(default=None)),
                        ("HttpPutResponseHopLimit", int, field(default=None)),
                        ("HttpEndpoint", str, field(default=None)),
                        ("HttpProtocolIpv6", str, field(default=None)),
                        ("InstanceMetadataTags", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "EnclaveOptions",
                make_dataclass(
                    "LaunchTemplateEnclaveOptionsRequest",
                    [("Enabled", bool, field(default=None))],
                ),
                field(default=None),
            ),
            (
                "InstanceRequirements",
                make_dataclass(
                    "InstanceRequirementsRequest",
                    [
                        (
                            "VCpuCount",
                            make_dataclass(
                                "VCpuCountRangeRequest",
                                [("Min", int), ("Max", int, field(default=None))],
                            ),
                        ),
                        (
                            "MemoryMiB",
                            make_dataclass(
                                "MemoryMiBRequest",
                                [("Min", int), ("Max", int, field(default=None))],
                            ),
                        ),
                        ("CpuManufacturers", List[str], field(default=None)),
                        (
                            "MemoryGiBPerVCpu",
                            make_dataclass(
                                "MemoryGiBPerVCpuRequest",
                                [
                                    ("Min", float, field(default=None)),
                                    ("Max", float, field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                        ("ExcludedInstanceTypes", List[str], field(default=None)),
                        ("InstanceGenerations", List[str], field(default=None)),
                        (
                            "SpotMaxPricePercentageOverLowestPrice",
                            int,
                            field(default=None),
                        ),
                        (
                            "OnDemandMaxPricePercentageOverLowestPrice",
                            int,
                            field(default=None),
                        ),
                        ("BareMetal", str, field(default=None)),
                        ("BurstablePerformance", str, field(default=None)),
                        ("RequireHibernateSupport", bool, field(default=None)),
                        (
                            "NetworkInterfaceCount",
                            make_dataclass(
                                "NetworkInterfaceCountRequest",
                                [
                                    ("Min", int, field(default=None)),
                                    ("Max", int, field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                        ("LocalStorage", str, field(default=None)),
                        ("LocalStorageTypes", List[str], field(default=None)),
                        (
                            "TotalLocalStorageGB",
                            make_dataclass(
                                "TotalLocalStorageGBRequest",
                                [
                                    ("Min", float, field(default=None)),
                                    ("Max", float, field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                        (
                            "BaselineEbsBandwidthMbps",
                            make_dataclass(
                                "BaselineEbsBandwidthMbpsRequest",
                                [
                                    ("Min", int, field(default=None)),
                                    ("Max", int, field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                        ("AcceleratorTypes", List[str], field(default=None)),
                        (
                            "AcceleratorCount",
                            make_dataclass(
                                "AcceleratorCountRequest",
                                [
                                    ("Min", int, field(default=None)),
                                    ("Max", int, field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                        ("AcceleratorManufacturers", List[str], field(default=None)),
                        ("AcceleratorNames", List[str], field(default=None)),
                        (
                            "AcceleratorTotalMemoryMiB",
                            make_dataclass(
                                "AcceleratorTotalMemoryMiBRequest",
                                [
                                    ("Min", int, field(default=None)),
                                    ("Max", int, field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                    ],
                ),
                field(default=None),
            ),
            (
                "PrivateDnsNameOptions",
                make_dataclass(
                    "LaunchTemplatePrivateDnsNameOptionsRequest",
                    [
                        ("HostnameType", str, field(default=None)),
                        ("EnableResourceNameDnsARecord", bool, field(default=None)),
                        ("EnableResourceNameDnsAAAARecord", bool, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "MaintenanceOptions",
                make_dataclass(
                    "LaunchTemplateInstanceMaintenanceOptionsRequest",
                    [("AutoRecovery", str, field(default=None))],
                ),
                field(default=None),
            ),
            ("DisableApiStop", bool, field(default=None)),
        ],
    ),
    name_prefix: str = None,
    resource_id: str = None,
    dry_run: bool = None,
    version_description: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
) -> Dict[str, Any]:
    """Creates a launch template.

    A launch template contains the parameters to launch an instance. When you launch an
    instance using RunInstances, you can specify a launch template instead of providing the launch parameters in the
    request. For more information, see Launch an instance from a launch template in the Amazon Elastic Compute Cloud
    User Guide. If you want to clone an existing launch template as the basis for creating a new launch template,
    you can use the Amazon EC2 console. The API, SDKs, and CLI do not support cloning a template. For more
    information, see Create a launch template from an existing launch template in the Amazon Elastic Compute Cloud
    User Guide.

    Args:
        name(str):
            An Idem name of the resource.
        name_prefix(str, Optional):
            Creates a unique name beginning with the specified prefix.

            .. Note::
                Either `name` or `name_prefix` should be specified.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.
        dry_run(bool, Optional):
            Checks whether you have the required permissions for the action, without actually making the
            request, and provides an error response. If you have the required permissions, the error
            response is DryRunOperation. Otherwise, it is UnauthorizedOperation. Defaults to None.
        version_description(str, Optional):
            A description for the first version of the launch template. Defaults to None.
        launch_template_data(Dict[str, Any]):
            The information for the launch template.
                * KernelId (str, Optional):
                    The ID of the kernel.  We recommend that you use PV-GRUB instead of kernels and RAM disks. For
                    more information, see User provided kernels in the Amazon Elastic Compute Cloud User Guide.
                * EbsOptimized (bool, Optional):
                    Indicates whether the instance is optimized for Amazon EBS I/O. This optimization provides
                    dedicated throughput to Amazon EBS and an optimized configuration stack to provide optimal
                    Amazon EBS I/O performance. This optimization isn't available with all instance types.
                    Additional usage charges apply when using an EBS-optimized instance.
                * IamInstanceProfile (dict[str, Any], Optional):
                    The name or Amazon Resource Name (ARN) of an IAM instance profile.
                        * Arn (str, Optional):
                            The Amazon Resource Name (ARN) of the instance profile.
                        * Name (str, Optional):
                            The name of the instance profile.
                * BlockDeviceMappings (list[dict[str, Any]], Optional):
                    The block device mapping.
                        * DeviceName (str, Optional):
                            The device name (for example, /dev/sdh or xvdh).
                        * VirtualName (str, Optional):
                            The virtual device name (ephemeralN). Instance store volumes are numbered starting from 0. An
                            instance type with 2 available instance store volumes can specify mappings for ephemeral0 and
                            ephemeral1. The number of available instance store volumes depends on the instance type. After
                            you connect to the instance, you must mount the volume.
                        * Ebs (dict[str, Any], Optional):
                            Parameters used to automatically set up EBS volumes when the instance is launched.
                                * Encrypted (bool, Optional):
                                    Indicates whether the EBS volume is encrypted. Encrypted volumes can only be attached to
                                    instances that support Amazon EBS encryption. If you are creating a volume from a snapshot, you
                                    can't specify an encryption value.
                                * DeleteOnTermination (bool, Optional):
                                    Indicates whether the EBS volume is deleted on instance termination.
                                * Iops (int, Optional):
                                    The number of I/O operations per second (IOPS). For gp3, io1, and io2 volumes, this represents
                                    the number of IOPS that are provisioned for the volume. For gp2 volumes, this represents the
                                    baseline performance of the volume and the rate at which the volume accumulates I/O credits for
                                    bursting. The following are the supported values for each volume type:    gp3: 3,000-16,000 IOPS
                                    io1: 100-64,000 IOPS    io2: 100-64,000 IOPS   For io1 and io2 volumes, we guarantee 64,000 IOPS
                                    only for Instances built on the Nitro System. Other instance families guarantee performance up
                                    to 32,000 IOPS. This parameter is supported for io1, io2, and gp3 volumes only. This parameter
                                    is not supported for gp2, st1, sc1, or standard volumes.
                                * KmsKeyId (str, Optional):
                                    The ARN of the symmetric Key Management Service (KMS) CMK used for encryption.
                                * SnapshotId (str, Optional):
                                    The ID of the snapshot.
                                * VolumeSize (int, Optional):
                                    The size of the volume, in GiBs. You must specify either a snapshot ID or a volume size. The
                                    following are the supported volumes sizes for each volume type: gp2 and gp3: 1-16,384    io1
                                    and io2: 4-16,384 st1 and sc1: 125-16,384 standard: 1-1,024
                                * VolumeType (str, Optional):
                                    The volume type. For more information, see Amazon EBS volume types in the Amazon Elastic Compute
                                    Cloud User Guide.
                                * Throughput (int, Optional):
                                    The throughput to provision for a gp3 volume, with a maximum of 1,000 MiB/s. Valid Range:
                                    Minimum value of 125. Maximum value of 1000.
                        * NoDevice (str, Optional):
                            To omit the device from the block device mapping, specify an empty string.
            * NetworkInterfaces (list[dict[str, Any]], Optional):
                One or more network interfaces. If you specify a network interface, you must specify any
                security groups and subnets as part of the network interface.

                * AssociateCarrierIpAddress (bool, Optional):
                        Associates a Carrier IP address with eth0 for a new network interface. Use this option when you
                        launch an instance in a Wavelength Zone and want to associate a Carrier IP address with the
                        network interface. For more information about Carrier IP addresses, see Carrier IP addresses in
                        the Wavelength Developer Guide.
                * AssociatePublicIpAddress (bool, Optional):
                        Associates a public IPv4 address with eth0 for a new network interface.

                * DeleteOnTermination (bool, Optional):
                        Indicates whether the network interface is deleted when the instance is terminated.

                * Description (str, Optional):
                    A description for the network interface.

                    * DeviceIndex (int, Optional):
                        The device index for the network interface attachment.

                    * Groups (list[str], Optional):
                        The IDs of one or more security groups.

                    * InterfaceType (str, Optional):
                        The type of network interface. To create an Elastic Fabric Adapter (EFA), specify efa. For more
                        information, see Elastic Fabric Adapter in the Amazon Elastic Compute Cloud User Guide. If you
                        are not creating an EFA, specify interface or omit this parameter. Valid values: interface | efa

                    * Ipv6AddressCount (int, Optional):
                        The number of IPv6 addresses to assign to a network interface. Amazon EC2 automatically selects
                        the IPv6 addresses from the subnet range. You can't use this option if specifying specific IPv6
                        addresses.

                    * Ipv6Addresses (list[dict[str, Any]], Optional):
                        One or more specific IPv6 addresses from the IPv6 CIDR block range of your subnet. You can't use
                        this option if you're specifying a number of IPv6 addresses.

                        * Ipv6Address (str, Optional):
                                The IPv6 address.

                    * NetworkInterfaceId (str, Optional):
                        The ID of the network interface.
                    * PrivateIpAddress (str, Optional):
                        The primary private IPv4 address of the network interface.
                    * PrivateIpAddresses (list[dict[str, Any]], Optional):
                        One or more private IPv4 addresses.
                            * Primary (bool, Optional):
                                Indicates whether the private IPv4 address is the primary private IPv4 address. Only one IPv4
                                address can be designated as primary.
                            * PrivateIpAddress (str, Optional):
                                The private IPv4 address.
                    * SecondaryPrivateIpAddressCount (int, Optional):
                        The number of secondary private IPv4 addresses to assign to a network interface.
                    * SubnetId (str, Optional):
                        The ID of the subnet for the network interface.
                    * NetworkCardIndex (int, Optional):
                        The index of the network card. Some instance types support multiple network cards. The primary
                        network interface must be assigned to network card index 0. The default is network card index 0.
                    * Ipv4Prefixes (list[dict[str, Any]], Optional):
                        One or more IPv4 prefixes to be assigned to the network interface. You cannot use this option if
                        you use the Ipv4PrefixCount option.

                        * Ipv4Prefix (str, Optional):
                                The IPv4 prefix. For information, see  Assigning prefixes to Amazon EC2 network interfaces in
                                the Amazon Elastic Compute Cloud User Guide.

                    * Ipv4PrefixCount (int, Optional):
                        The number of IPv4 prefixes to be automatically assigned to the network interface. You cannot
                        use this option if you use the Ipv4Prefix option.
                    * Ipv6Prefixes (list[dict[str, Any]], Optional):
                        One or more IPv6 prefixes to be assigned to the network interface. You cannot use this option if
                        you use the Ipv6PrefixCount option.

                        * Ipv6Prefix (str, Optional):
                                The IPv6 prefix.

                    * Ipv6PrefixCount (int, Optional):
                        The number of IPv6 prefixes to be automatically assigned to the network interface. You cannot
                        use this option if you use the Ipv6Prefix option.
            * ImageId (str, Optional):
                The ID of the AMI.
            * InstanceType (str, Optional):
                The instance type. For more information, see Instance types in the Amazon Elastic Compute Cloud
                User Guide. If you specify InstanceType, you can't specify InstanceRequirements.
            * KeyName (str, Optional):
                The name of the key pair. You can create a key pair using CreateKeyPair or ImportKeyPair.  If
                you do not specify a key pair, you can't connect to the instance unless you choose an AMI that
                is configured to allow users another way to log in.
            * Monitoring (dict[str, Any], Optional):
                The monitoring for the instance.
                    * Enabled (bool, Optional):
                        Specify true to enable detailed monitoring. Otherwise, basic monitoring is enabled.
            * Placement (dict[str, Any], Optional):
                The placement for the instance.
                    * AvailabilityZone (str, Optional):
                        The Availability Zone for the instance.
                    * Affinity (str, Optional):
                        The affinity setting for an instance on a Dedicated Host.
                    * GroupName (str, Optional):
                        The name of the placement group for the instance.
                    * HostId (str, Optional):
                        The ID of the Dedicated Host for the instance.
                    * Tenancy (str, Optional):
                        The tenancy of the instance (if the instance is running in a VPC). An instance with a tenancy of
                        dedicated runs on single-tenant hardware.
                    * SpreadDomain (str, Optional):
                        Reserved for future use.
                    * HostResourceGroupArn (str, Optional):
                        The ARN of the host resource group in which to launch the instances. If you specify a host
                        resource group ARN, omit the Tenancy parameter or set it to host.
                    * PartitionNumber (int, Optional):
                        The number of the partition the instance should launch in. Valid only if the placement group
                        strategy is set to partition.
            * RamDiskId (str, Optional):
                The ID of the RAM disk.  We recommend that you use PV-GRUB instead of kernels and RAM disks. For
                more information, see User provided kernels in the Amazon Elastic Compute Cloud User Guide.
            * DisableApiTermination (bool, Optional):
                If you set this parameter to true, you can't terminate the instance using the Amazon EC2
                console, CLI, or API; otherwise, you can. To change this attribute after launch, use
                ModifyInstanceAttribute. Alternatively, if you set InstanceInitiatedShutdownBehavior to
                terminate, you can terminate the instance by running the shutdown command from the instance.
            * InstanceInitiatedShutdownBehavior (str, Optional):
                Indicates whether an instance stops or terminates when you initiate shutdown from the instance
                (using the operating system command for system shutdown). Default: stop
            * UserData (str, Optional):
                The user data to make available to the instance. You must provide base64-encoded str. User data
                is limited to 16 KB. For more information, see Run commands on your Linux instance at launch
                (Linux) or Work with instance user data (Windows) in the Amazon Elastic Compute Cloud User
                Guide. If you are creating the launch template for use with Batch, the user data must be
                provided in the  MIME multi-part archive format. For more information, see Amazon EC2 user data
                in launch templates in the Batch User Guide.
            * TagSpecifications (list[dict[str, Any]], Optional):
                The tags to apply to the resources that are created during instance launch. You can specify tags
                for the following resources only:   Instances   Volumes   Elastic graphics   Spot Instance
                requests   Network interfaces   To tag a resource after it has been created, see CreateTags.  To
                tag the launch template itself, you must use the TagSpecification parameter.

                * ResourceType (str, Optional):
                        The type of resource to tag. The Valid Values are all the resource types that can be tagged.
                        However, when creating a launch template, you can specify tags for the following resource types
                        only: instance | volume | elastic-gpu | network-interface | spot-instances-request  To tag a
                        resource after it has been created, see CreateTags.
                * Tags (list[dict[str, Any]], Optional):
                        The tags to apply to the resource.

                        * Key (str, Optional):
                                The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                                characters. May not begin with aws:.

                        * Value (str, Optional):
                                The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                                Unicode characters.
            * ElasticGpuSpecifications (list[dict[str, Any]], Optional):
                An elastic GPU to associate with the instance.

                * Type (str):
                    The type of Elastic Graphics accelerator. For more information about the values to specify for
                    Type, see Elastic Graphics Basics, specifically the Elastic Graphics accelerator column, in the
                    Amazon Elastic Compute Cloud User Guide for Windows Instances.

            * ElasticInferenceAccelerators (list[dict[str, Any]], Optional):
                The elastic inference accelerator for the instance.

                * Type (str):
                    The type of elastic inference accelerator. The possible values are eia1.medium, eia1.large, and
                    eia1.xlarge.

                * Count (int, Optional):
                    The number of elastic inference accelerators to attach to the instance.

                    Default: 1

            * SecurityGroupIds (list[str], Optional):
                One or more security group IDs. You can create a security group using CreateSecurityGroup. You
                cannot specify both a security group ID and security name in the same request.
            * SecurityGroups (list[str], Optional):
                One or more security group names. For a nondefault VPC, you must use security group IDs instead.
                You cannot specify both a security group ID and security name in the same request.
            * InstanceMarketOptions (dict[str, Any], Optional):
                The market (purchasing) option for the instances.

                * MarketType (str, Optional):
                    The market type.

                * SpotOptions (Dict[str, Any], Optional):
                    The options for Spot Instances.

                    * MaxPrice (str, Optional):
                        The maximum hourly price you're willing to pay for the Spot Instances. We do not recommend using
                        this parameter because it can lead to increased interruptions. If you do not specify this
                        parameter, you will pay the current Spot price.  If you specify a maximum price, your Spot
                        Instances will be interrupted more frequently than if you do not specify this parameter.

                    * SpotInstanceType (str, Optional):
                        The Spot Instance request type.

                    * BlockDurationMinutes (int, Optional):
                        Deprecated.

                    * ValidUntil (datetime, Optional):
                        The end date of the request, in UTC format (YYYY-MM-DDTHH:MM:SSZ). Supported only for persistent
                        requests.   For a persistent request, the request remains active until the ValidUntil date and
                        time is reached. Otherwise, the request remains active until you cancel it.   For a one-time
                        request, ValidUntil is not supported. The request remains active until all instances launch or
                        you cancel the request.   Default: 7 days from the current date

                    * InstanceInterruptionBehavior (str, Optional):
                        The behavior when a Spot Instance is interrupted. The default is terminate.

            * CreditSpecification (dict[str, Any], Optional):
                The credit option for CPU usage of the instance. Valid only for T instances.
                * CpuCredits (str):
                    The credit option for CPU usage of a T instance. Valid values: standard | unlimited
            * CpuOptions (dict[str, Any], Optional):
                The CPU options for the instance. For more information, see Optimizing CPU Options in the Amazon
            Elastic Compute Cloud User Guide.
                * CoreCount (int, Optional):
                    The number of CPU cores for the instance.
                * ThreadsPerCore (int, Optional):
                    The number of threads per CPU core. To disable multithreading for the instance, specify a value
                    of 1. Otherwise, specify the default value of 2.
            * CapacityReservationSpecification (dict[str, Any], Optional):
                The Capacity Reservation targeting option. If you do not specify this parameter, the instance's
                Capacity Reservation preference defaults to open, which enables it to run in any open Capacity
                Reservation that has matching attributes (instance type, platform, Availability Zone).
                * CapacityReservationPreference (str, Optional):
                    Indicates the instance's Capacity Reservation preferences. Possible preferences include:
                     open - The instance can run in any open Capacity Reservation that has matching attributes (instance
                            type, platform, Availability Zone).
                     none - The instance avoids running in a Capacity Reservation even if one is available.
                            The instance runs in On-Demand capacity.
                * CapacityReservationTarget (dict[str, Any], Optional):
                    Information about the target Capacity Reservation or Capacity Reservation group.
                    * CapacityReservationId (str, Optional):
                        The ID of the Capacity Reservation in which to run the instance.
                    * CapacityReservationResourceGroupArn (str, Optional):
                        The ARN of the Capacity Reservation resource group in which to run the instance.
            * LicenseSpecifications (list[dict[str, Any]], Optional):
                The license configurations.
                * LicenseConfigurationArn (str, Optional):
                    The Amazon Resource Name (ARN) of the license configuration.
            * HibernationOptions (dict[str, Any], Optional):
                Indicates whether an instance is enabled for hibernation. This parameter is valid only if the
                instance meets the hibernation prerequisites. For more information, see Hibernate your instance
                in the Amazon Elastic Compute Cloud User Guide.
                * Configured (bool, Optional):
                    If you set this parameter to true, the instance is enabled for hibernation. Default: false
            * MetadataOptions (dict[str, Any], Optional):
                The metadata options for the instance. For more information, see Instance metadata and user data
                in the Amazon Elastic Compute Cloud User Guide.
                * HttpTokens (str, Optional):
                    The state of token usage for your instance metadata requests. If the parameter is not specified
                    in the request, the default state is Optional. If the state is Optional, you can choose to
                    retrieve instance metadata with or without a signed token header on your request. If you
                    retrieve the IAM role credentials without a token, the version 1.0 role credentials are
                    returned. If you retrieve the IAM role credentials using a valid signed token, the version 2.0
                    role credentials are returned. If the state is required, you must send a signed token header
                    with any instance metadata retrieval requests. In this state, retrieving the IAM role
                    credentials always returns the version 2.0 credentials; the version 1.0 credentials are not
                    available.
                * HttpPutResponseHopLimit (int, Optional):
                    The desired HTTP PUT response hop limit for instance metadata requests. The larger the number,
                    the further instance metadata requests can travel. Default: 1  Possible values: Integers from 1
                    to 64
                * HttpEndpoint (str, Optional):
                    Enables or disables the HTTP metadata endpoint on your instances. If the parameter is not
                    specified, the default state is enabled.  If you specify a value of disabled, you will not be
                    able to access your instance metadata.
                * HttpProtocolIpv6 (str, Optional):
                    Enables or disables the IPv6 endpoint for the instance metadata service. Default: disabled
                * InstanceMetadataTags (str, Optional):
                    Set to enabled to allow access to instance tags from the instance metadata. Set to disabled to
                    turn off access to instance tags from the instance metadata. For more information, see Work with
                    instance tags using the instance metadata. Default: disabled
            * EnclaveOptions (dict[str, Any], Optional):
                Indicates whether the instance is enabled for Amazon Web Services Nitro Enclaves. For more
                information, see  What is Amazon Web Services Nitro Enclaves? in the Amazon Web Services Nitro
                Enclaves User Guide. You can't enable Amazon Web Services Nitro Enclaves and hibernation on the
                same instance.
                * Enabled (bool, Optional):
                    To enable the instance for Amazon Web Services Nitro Enclaves, set this parameter to true.
            * InstanceRequirements (dict[str, Any], Optional):
                The attributes for the instance types. When you specify instance attributes, Amazon EC2 will
                identify instance types with these attributes. If you specify InstanceRequirements, you can't
                specify InstanceType.
                * VCpuCount (dict[str, Any]):
                    The minimum and maximum number of vCPUs.
                    * Min (int):
                        The minimum number of vCPUs. To specify no minimum limit, specify 0.
                    * Max (int, Optional): The maximum number of vCPUs. To specify no maximum limit, omit this parameter.
                * MemoryMiB (dict[str, Any]):
                    The minimum and maximum amount of memory, in MiB.
                    * Min (int):
                        The minimum amount of memory, in MiB. To specify no minimum limit, specify 0.
                    * Max (int, Optional):
                        The maximum amount of memory, in MiB. To specify no maximum limit, omit this parameter.
                * CpuManufacturers (list[str], Optional):
                    The CPU manufacturers to include. For instance types with Intel CPUs, specify intel. For
                    instance types with AMD CPUs, specify amd. For instance types with Amazon Web Services CPUs,
                    specify amazon-web-services. Don't confuse the CPU manufacturer with the CPU architecture.
                    Instances will be launched with a compatible CPU architecture based on the Amazon Machine Image
                    (AMI) that you specify in your launch template. Default: Any manufacturer
                * MemoryGiBPerVCpu (dict[str, Any], Optional):
                    The minimum and maximum amount of memory per vCPU, in GiB. Default: No minimum or maximum limits
                    * Min (float, Optional):
                        The minimum amount of memory per vCPU, in GiB. To specify no minimum limit, omit this parameter.
                    * Max (float, Optional):
                        The maximum amount of memory per vCPU, in GiB. To specify no maximum limit, omit this parameter.
                * ExcludedInstanceTypes (list[str], Optional):
                    The instance types to exclude. You can use strings with one or more wild cards, represented by
                    an asterisk (*), to exclude an instance family, type, size, or generation. The following are
                    examples: m5.8xlarge, c5*.*, m5a.*, r*, *3*. For example, if you specify c5*,Amazon EC2 will
                    exclude the entire C5 instance family, which includes all C5a and C5n instance types. If you
                    specify m5a.*, Amazon EC2 will exclude all the M5a instance types, but not the M5n instance
                    types. Default: No excluded instance types
                * InstanceGenerations (list[str], Optional):
                    Indicates whether current or previous generation instance types are included. The current
                    generation instance types are recommended for use. Current generation instance types are
                    typically the latest two to three generations in each instance family. For more information, see
                    Instance types in the Amazon EC2 User Guide. For current generation instance types, specify
                    current. For previous generation instance types, specify previous. Default: Current and previous
                    generation instance types
                * SpotMaxPricePercentageOverLowestPrice (int, Optional):
                    The price protection threshold for Spot Instance. This is the maximum you’ll pay for an Spot
                    Instance, expressed as a percentage above the least expensive current generation M, C, or R
                    instance type with your specified attributes. When Amazon EC2 selects instance types with your
                    attributes, it excludes instance types priced above your threshold. The parameter accepts an
                    integer, which Amazon EC2 interprets as a percentage. To turn off price protection, specify a
                    high value, such as 999999. This parameter is not supported for GetSpotPlacementScores and
                    GetInstanceTypesFromInstanceRequirements.  If you set TargetCapacityUnitType to vcpu or memory-
                    mib, the price protection threshold is applied based on the per-vCPU or per-memory price instead
                    of the per-instance price.  Default: 100
                * OnDemandMaxPricePercentageOverLowestPrice (int, Optional):
                    The price protection threshold for On-Demand Instances. This is the maximum you’ll pay for an
                On-Demand Instance, expressed as a percentage above the least expensive current generation M, C,
                or R instance type with your specified attributes. When Amazon EC2 selects instance types with
                your attributes, it excludes instance types priced above your threshold. The parameter accepts
                an integer, which Amazon EC2 interprets as a percentage. To turn off price protection, specify a
                high value, such as 999999. This parameter is not supported for GetSpotPlacementScores and
                GetInstanceTypesFromInstanceRequirements.  If you set TargetCapacityUnitType to vcpu or memory-
                mib, the price protection threshold is applied based on the per-vCPU or per-memory price instead
                of the per-instance price.  Default: 20
                * BareMetal (str, Optional):
                    Indicates whether bare metal instance types must be included, excluded, or required.   To
                    include bare metal instance types, specify included.   To require only bare metal instance
                    types, specify required.   To exclude bare metal instance types, specify excluded.   Default:
                    excluded
                * BurstablePerformance (str, Optional):
                    Indicates whether burstable performance T instance types are included, excluded, or required.
                    For more information, see Burstable performance instances.   To include burstable performance
                    instance types, specify included.   To require only burstable performance instance types,
                    specify required.   To exclude burstable performance instance types, specify excluded.
                    Default: excluded
                * RequireHibernateSupport (bool, Optional):
                    Indicates whether instance types must support hibernation for On-Demand Instances. This
                    parameter is not supported for GetSpotPlacementScores. Default: false
                * NetworkInterfaceCount (dict[str, Any], Optional):
                    The minimum and maximum number of network interfaces. Default: No minimum or maximum limits

                    * Min (int, Optional):
                        The minimum number of network interfaces. To specify no minimum limit, omit this parameter.

                    * Max (int, Optional):
                        The maximum number of network interfaces. To specify no maximum limit, omit this parameter.

                * LocalStorage (str, Optional):
                    Indicates whether instance types with instance store volumes are included, excluded, or
                    required. For more information, Amazon EC2 instance store in the Amazon EC2 User Guide.   To
                    include instance types with instance store volumes, specify included.   To require only instance
                    types with instance store volumes, specify required.   To exclude instance types with instance
                    store volumes, specify excluded.   Default: included
                * LocalStorageTypes (list[str], Optional):
                    The type of local storage that is required. For instance types with hard disk drive (HDD)
                    storage, specify hdd. For instance types with solid state drive (SSD) storage, specify ssd.
                    Default: hdd and ssd
                * TotalLocalStorageGB (dict[str, Any], Optional):
                    The minimum and maximum amount of total local storage, in GB. Default: No minimum or maximum
                    limits

                    * Min (float, Optional):
                        The minimum amount of total local storage, in GB. To specify no minimum limit, omit this
                        parameter.

                    * Max (float, Optional):
                        The maximum amount of total local storage, in GB. To specify no maximum limit, omit this
                        parameter.

                * BaselineEbsBandwidthMbps (dict[str, Any], Optional):
                    The minimum and maximum baseline bandwidth to Amazon EBS, in Mbps. For more information, see
                    Amazon EBS–optimized instances in the Amazon EC2 User Guide. Default: No minimum or maximum
                    limits

                    * Min (int, Optional):
                        The minimum baseline bandwidth, in Mbps. To specify no minimum limit, omit this parameter.

                    * Max (int, Optional):
                        The maximum baseline bandwidth, in Mbps. To specify no maximum limit, omit this parameter.

                * AcceleratorTypes (list[str], Optional):
                    The accelerator types that must be on the instance type.   To include instance types with GPU
                    hardware, specify gpu.   To include instance types with FPGA hardware, specify fpga.   To
                    include instance types with inference hardware, specify inference.   Default: Any accelerator
                    type
                * AcceleratorCount (dict[str, Any], Optional):
                    The minimum and maximum number of accelerators (GPUs, FPGAs, or Amazon Web Services Inferentia
                    chips) on an instance. To exclude accelerator-enabled instance types, set Max to 0. Default: No
                    minimum or maximum limits

                    * Min (int, Optional):
                        The minimum number of accelerators. To specify no minimum limit, omit this parameter.

                    * Max (int, Optional):
                        The maximum number of accelerators. To specify no maximum limit, omit this parameter. To exclude
                        accelerator-enabled instance types, set Max to 0.

                * AcceleratorManufacturers (list[str], Optional):
                    Indicates whether instance types must have accelerators by specific manufacturers.   For
                    instance types with NVIDIA devices, specify nvidia.   For instance types with AMD devices,
                    specify amd.   For instance types with Amazon Web Services devices, specify amazon-web-services.
                    For instance types with Xilinx devices, specify xilinx.   Default: Any manufacturer

                * AcceleratorNames (list[str], Optional):
                    The accelerators that must be on the instance type.   For instance types with NVIDIA A100 GPUs,
                    specify a100.   For instance types with NVIDIA V100 GPUs, specify v100.   For instance types
                    with NVIDIA K80 GPUs, specify k80.   For instance types with NVIDIA T4 GPUs, specify t4.   For
                    instance types with NVIDIA M60 GPUs, specify m60.   For instance types with AMD Radeon Pro V520
                    GPUs, specify radeon-pro-v520.   For instance types with Xilinx VU9P FPGAs, specify  vu9p.   For
                    instance types with Amazon Web Services Inferentia chips, specify inferentia.   For instance
                    types with NVIDIA GRID K520 GPUs, specify k520.   Default: Any accelerator

                * AcceleratorTotalMemoryMiB (dict[str, Any], Optional):
                    The minimum and maximum amount of total accelerator memory, in MiB. Default: No minimum or
                    maximum limits

                    * Min (int, Optional):
                        The minimum amount of accelerator memory, in MiB. To specify no minimum limit, omit this
                        parameter.

                    * Max (int, Optional):
                        The maximum amount of accelerator memory, in MiB. To specify no maximum limit, omit this
                        parameter.

            * PrivateDnsNameOptions (dict[str, Any], Optional):
                The options for the instance hostname. The default values are inherited from the subnet.

                * HostnameType (str, Optional):
                    The type of hostname for Amazon EC2 instances. For IPv4 only subnets, an instance DNS name must
                    be based on the instance IPv4 address. For IPv6 native subnets, an instance DNS name must be
                    based on the instance ID. For dual-stack subnets, you can specify whether DNS names use the
                    instance IPv4 address or the instance ID.

                * EnableResourceNameDnsARecord (bool, Optional):
                    Indicates whether to respond to DNS queries for instance hostnames with DNS A records.

                * EnableResourceNameDnsAAAARecord (bool, Optional):
                    Indicates whether to respond to DNS queries for instance hostnames with DNS AAAA records.

            * MaintenanceOptions (dict[str, Any], Optional):
                The maintenance options for the instance.

                * AutoRecovery (str, Optional):
                    Disables the automatic recovery behavior of your instance or sets it to default. For more
                    information, see Simplified automatic recovery.

            * DisableApiStop (bool, Optional):
                Indicates whether to enable the instance for stop protection. For more information, see Stop
                Protection.

        tags(Dict or List, Optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the VPC.
            Each tag consists of a key name and an associated value. Defaults to None.
            * Key (str, Optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * Value(str, Optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.

    Request Syntax:
       .. code-block:: sls

          [launch-template-resource-id]:
              aws.ec2.launch_template.present:
                - name: string
                - name_prefix: string
                - dry_run: boolean
                - version_description: string
                - launch_template_data: dict
                - tags: dict or list

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            my_resource:
                aws.ec2.launch_template.present:
                  - name: Testversions
                  - resource_id: lt-08467ebddfaed30ad
                  - launch_template_name: Testversions
                  - create_time: '2022-11-08T08:55:48Z'
                  - tags:
                      Name: Testversions
                  - created_by: arn:aws:iam::746014882121:user/photonautomation
                  - default_version_number: 1
                  - latest_version_number: 4
                  - version_description: latest
                  - launch_template_data:
                      DisableApiTermination: false
                      HibernationOptions:
                        Configured: false
                      IamInstanceProfile:
                        Arn: arn:aws:iam::746014882121:instance-profile/bootstrap-idem-test-rs-terraform-eks
                      ImageId: ami-074b543ee210b32e7
                      InstanceInitiatedShutdownBehavior: terminate
                      InstanceRequirements:
                        MemoryMiB:
                          Max: 60
                          Min: 60
                        VCpuCount:
                          Max: 1
                          Min: 1
                      MaintenanceOptions:
                        AutoRecovery: default
                      MetadataOptions:
                        HttpEndpoint: enabled
                        HttpTokens: optional
                        InstanceMetadataTags: disabled
                      Monitoring:
                        Enabled: false
                      NetworkInterfaces:
                      - DeviceIndex: 0
                        Groups:
                        - sg-0ab44a7b6b855692f
                        SubnetId: subnet-0e15343bc21685b42
                      Placement:
                        Tenancy: dedicated
                      PrivateDnsNameOptions:
                        HostnameType: ip-name
                      TagSpecifications:
                      - ResourceType: instance
                        Tags:
                        - Key: name
                          Value: my_resource
                      - ResourceType: volume
                        Tags:
                        - Key: name
                          Value: bootstrap-idem-test-rs-fin
                      UserData: bHM=
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if isinstance(tags, Dict):
        tags = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
    if resource_id:
        before = await hub.exec.aws.ec2.launch_template.get(
            ctx, name=name, resource_id=resource_id
        )
    if before:
        if not before.get("result") and not before.get("ret"):
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = copy.deepcopy(before["ret"])
        result["old_state"]["name_prefix"] = name_prefix
        plan_state = copy.deepcopy(result["old_state"])
        old_tags = result["old_state"].get("tags")
        old_tags_list = hub.tool.aws.tag_utils.convert_tag_dict_to_list(old_tags)
        # Update tags
        if not hub.tool.aws.state_comparison_utils.are_lists_identical(
            tags, old_tags_list
        ):
            update_ret = await hub.tool.aws.ec2.tag.update_tags(
                ctx,
                resource_id=resource_id,
                old_tags=old_tags,
                new_tags=tags,
            )
            result["result"] = result["result"] and update_ret["result"]
            result["comment"] = update_ret["comment"]
            resource_updated = resource_updated or bool(update_ret["result"])
            if ctx.get("test", False) and update_ret["result"]:
                plan_state["tags"] = update_ret["ret"]
        update_version_ret = (
            await hub.tool.aws.ec2.launch_template.update_launch_template_version(
                ctx=ctx,
                launch_template_id=resource_id,
                old_launch_template_data=result["old_state"].get(
                    "launch_template_data"
                ),
                new_launch_template_data=launch_template_data,
            )
        )
        if ctx.get("test", False) and update_version_ret["ret"]:
            plan_state["launch_template_data"] = update_version_ret["ret"]
        result["comment"] = result["comment"] + update_version_ret["comment"]
        resource_updated = resource_updated or bool(update_version_ret["ret"])
        result["result"] = result["result"] and update_version_ret["result"]
        if ctx.get("test", False) and resource_updated:
            result["new_state"] = plan_state
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.ec2.launch_template", name=name
            )
            return result
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "name_prefix": name_prefix,
                    "dry_run": dry_run,
                    "version_description": version_description,
                    "tags": tags,
                    "launch_template_data": launch_template_data,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.launch_template", name=name
            )
            return result
        ret = await hub.exec.boto3.client.ec2.create_launch_template(
            ctx,
            ClientToken=name,
            DryRun=dry_run,
            LaunchTemplateName=name,
            VersionDescription=version_description,
            LaunchTemplateData=hub.tool.aws.ec2.launch_template.pre_process_launch_template_data(
                chunk=launch_template_data
            ),
            TagSpecifications=(
                [
                    {
                        "ResourceType": "launch-template",
                        "Tags": tags,
                    }
                ]
                if tags
                else None
            ),
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["LaunchTemplate"]["LaunchTemplateId"]
        # This makes sure the created launch template is saved to esm regardless if the subsequent update call fails or not.
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ec2.launch_template", name=name
        )
    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not result["old_state"]) or resource_updated:
        after = await hub.exec.aws.ec2.launch_template.get(
            ctx=ctx, name=name, resource_id=resource_id
        )

        if not after["result"]:
            result["result"] = False
            result["comment"] = after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
        result["new_state"]["name_prefix"] = name_prefix
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str,
) -> Dict[str, Any]:
    """Deletes a launch template.

    Deleting a launch template deletes all of its versions.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str):
            An identifier of the resource in the provider.

    Request Syntax:
       .. code-block:: sls

          [launch_template-id]:
              aws.ec2.launch_template.absent:
                - name: 'string'
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.ec2.launch_template.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.launch_template", name=name
        )
        return result
    before = await hub.exec.aws.ec2.launch_template.get(
        ctx=ctx, name=name, resource_id=resource_id
    )

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    elif not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.launch_template", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.ec2.launch_template", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.ec2.delete_launch_template(
            ctx, LaunchTemplateId=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.ec2.launch_template", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes one or more launch templates.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.ec2.launch_template
    """
    result = {}
    ret = await hub.exec.aws.ec2.launch_template.list(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe launch_template {ret['comment']}")
        return {}

    for resource in ret["ret"]:
        result[resource["resource_id"]] = {
            "aws.ec2.launch_template.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result



"""Util functions for converting raw attributes to present form and vice versa."""
import json
from collections import OrderedDict
from json import JSONDecodeError
from typing import Any
from typing import Dict


def convert_raw_subscription_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    """Util functions to convert raw resource state to present input format for SNS topic_subscription."""
    raw_attributes = raw_resource.get("Attributes")
    resource_id = raw_attributes.get("SubscriptionArn")
    resource_parameters = OrderedDict(
        {
            "TopicArn": "topic_arn",
            "Protocol": "protocol",
            "Endpoint": "endpoint",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource.get("Attributes"):
            resource_translated[parameter_present] = raw_resource.get("Attributes").get(
                parameter_raw
            )

    attribute_params = [
        "DeliveryPolicy",
        "FilterPolicy",
        "RawMessageDelivery",
        "RedrivePolicy",
    ]

    attributes = {}
    for param in attribute_params:
        value = raw_attributes.get(param, None)
        if value:
            attributes[param] = standardise_json(hub, value)
    resource_translated["attributes"] = attributes

    return resource_translated


def convert_raw_topic_to_present(
    hub,
    raw_resource: Dict[str, Any],
    raw_resource_tags: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    """Util functions to convert raw resource state to present input format for SNS topic."""
    raw_attributes = raw_resource.get("Attributes")
    resource_id = raw_attributes.get("TopicArn")
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    attribute_params = [
        "DeliveryPolicy",
        "DisplayName",
        "FifoTopic",
        "Policy",
        "KmsMasterKeyId",
        "ContentBasedDeduplication",
    ]

    attributes = {}
    for param in attribute_params:
        if raw_attributes.get(param):
            attributes[param] = raw_attributes.get(param)
    resource_translated["attributes"] = attributes

    if raw_resource_tags.get("ret") and raw_resource_tags.get("ret").get("Tags"):
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource_tags.get("ret").get("Tags")
        )

    return resource_translated


def convert_raw_topic_policy_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    """Util functions to convert raw resource state to present input format for SNS topic_policy."""
    raw_attributes = raw_resource["ret"].get("Attributes")
    resource_id = raw_attributes.get("TopicArn") + "-policy"
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "topic_arn": raw_attributes.get("TopicArn"),
        "policy": standardise_json(hub, raw_attributes.get("Policy")),
    }

    return resource_translated


def standardise_json(hub, value: str):
    """Utils function for standardising the json string format for SNS resources."""
    if value is None:
        return None

    if isinstance(value, str):
        try:
            json_dict = json.loads(value)
        except JSONDecodeError:
            json_dict = value
    else:
        json_dict = value
    return json.dumps(json_dict, separators=(", ", ": "))



"""Exec module for managing EKS cluster."""


async def get(hub, ctx, name, resource_id: str):
    """Provides details about a specific cluster.

     Args:
         name (str):
             The name of the Idem state.

         resource_id (str):
             EKS cluster id to identify the resource.

    Returns:
         Dict[str, Any]:
             Returns ami in present format

     Examples:
         Calling this exec module function from the cli

         .. code-block:: bash

             idem exec aws.eks.cluster.get name="my_resource"

         Using in a state:

         .. code-block:: yaml

             my_unmanaged_resource:
               exec.run:
                 - path: aws.eks.cluster.get
                 - kwargs:
                     name: my_resource
    """
    result = dict(comment=[], ret=None, result=True)

    ret = await hub.exec.boto3.client.eks.describe_cluster(ctx, name=resource_id)

    if not ret["result"]:
        if "" in str(ret["comment"]) or "ResourceNotFoundException" in str(
            ret["comment"]
        ):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.eks.cluster", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        # Failed to retrieve
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    result["ret"] = hub.tool.aws.eks.conversion_utils.convert_raw_cluster_to_present(
        raw_resource=ret["ret"]["cluster"],
        idem_resource_name=name,
    )

    return result



"""Exec module for managing Amazon API Gateway Resources."""


async def get(hub, ctx, name: str, resource_id: str, rest_api_id: str):
    """Get an API Gateway Resource from AWS.

    Args:
        name (str): An Idem name of the API Gateway Resource.

        resource_id (str): AWS Resource id of the associated Rest API.

        rest_api_id (str): AWS rest_api id of the associated RestApi.
    """

    result = dict(comment=[], ret=None, result=True)
    before = await hub.exec.boto3.client.apigateway.get_resource(
        ctx, resourceId=resource_id, restApiId=rest_api_id
    )

    if not before["result"]:
        if "NotFoundException" in str(before["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.api.gateway", name=name
                )
            )
            result["comment"] += list(before["comment"])
            return result
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    if before["ret"]:
        result["ret"] = before["ret"]

    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

STATE_NAME = "aws.docdb.db_cluster_parameter_group"


async def search_raw(hub, ctx, filters: List = None, resource_id: str = None) -> Dict:
    """
    Fetch one or more docdb db_cluster_parameter_group objects in its raw form.

    Args:
        filters(list, Optional): list of filters for boto3 call
        resource_id(str, Optional): Document DB Cluster Parameter Group Name

    Returns:
        .. code-block:: python
        {"result": True|False, "comment": A message List, "ret": Dict}
    """

    result = dict(comment=[], ret=None, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["result"] = False
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )

    ret = await hub.exec.boto3.client.docdb.describe_db_cluster_parameter_groups(
        ctx,
        DBClusterParameterGroupName=resource_id if resource_id else None,
        Filters=boto3_filter,
    )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result


async def update(
    hub,
    ctx,
    name: str,
    resource_id: str,
    desired_state: Dict,
    is_created: bool,
) -> Dict[str, Any]:
    result = dict(comment=[], old_state=None, new_state=None, result=True)
    if ctx.get("test", False) and is_created:
        current_state = desired_state
    else:
        current_state_result = await hub.exec.aws.docdb.db_cluster_parameter_group.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not (
            current_state_result
            and current_state_result["result"]
            and current_state_result["ret"]
        ):
            result["result"] = False
            result["comment"] = current_state_result["comment"]
            return result
        current_state = current_state_result["ret"]
    if not is_created:
        result["old_state"] = current_state
    resource_arn = current_state.get("db_cluster_parameter_group_arn")
    params_to_modify = {}
    plan_state = copy.deepcopy(current_state)

    if None not in (
        desired_state,
        desired_state["parameters"],
    ) and not hub.tool.aws.state_comparison_utils.are_lists_identical(
        current_state["parameters"], desired_state["parameters"]
    ):
        params_to_modify["Parameters"] = desired_state["parameters"]
        plan_state["parameters"] = desired_state["parameters"]

    if params_to_modify:
        if ctx.get("test", False):
            if not is_created:
                result["comment"].append(
                    hub.tool.aws.comment_utils.would_update_comment(
                        resource_type=STATE_NAME, name=name
                    )
                )
        else:
            modify_db_cluster_parameter_group_result = (
                await hub.exec.boto3.client.docdb.modify_db_cluster_parameter_group(
                    ctx, DBClusterParameterGroupName=resource_id, **params_to_modify
                )
            )
            if not modify_db_cluster_parameter_group_result["result"]:
                result["comment"] = modify_db_cluster_parameter_group_result["comment"]
                result["result"] = False
                return result
            if not is_created:
                result["comment"].append(
                    hub.tool.aws.comment_utils.update_comment(
                        resource_type=STATE_NAME, name=name
                    )
                )

    old_tags = current_state.get("tags")
    new_tags = desired_state.get("tags")
    if new_tags is not None and new_tags != old_tags:
        update_tags_ret = await hub.tool.aws.docdb.tag.update_tags(
            ctx=ctx, resource_arn=resource_arn, old_tags=old_tags, new_tags=new_tags
        )
        if not update_tags_ret["result"]:
            result["comment"] = update_tags_ret["comment"]
            result["result"] = False
            return result
        if not is_created:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type=STATE_NAME, name=name
            )
            result["comment"] += update_tags_ret["comment"]

        if ctx.get("test", False) and update_tags_ret["ret"] is not None:
            plan_state["tags"] = update_tags_ret["ret"]
            if not is_created:
                result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                    resource_type=STATE_NAME, name=resource_id
                )
                result["comment"] += update_tags_ret["comment"]

    # set new_state
    if ctx.get("test", False):
        result["new_state"] = plan_state
    else:
        current_db_cluster_parameter_group_result = (
            await hub.exec.aws.docdb.db_cluster_parameter_group.get(
                ctx=ctx, name=name, resource_id=resource_id
            )
        )
        if not (
            current_db_cluster_parameter_group_result
            and current_db_cluster_parameter_group_result["result"]
            and current_db_cluster_parameter_group_result["ret"]
        ):
            result["result"] = False
            result["comment"] += current_db_cluster_parameter_group_result["comment"]
            return result
        result["new_state"] = current_db_cluster_parameter_group_result["ret"]

    return result


def convert_raw_db_cluster_parameter_group_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str,
    tags: List = None,
    parameters: List[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    raw_resource["Parameters"] = parameters
    resource_parameters = OrderedDict(
        {
            "DBClusterParameterGroupName": "resource_id",
            "DBParameterGroupFamily": "db_parameter_group_family",
            "Description": "description",
            "Parameters": "parameters",
            "DBClusterParameterGroupArn": "db_cluster_parameter_group_arn",
        }
    )
    resource_translated = {"name": raw_resource.get("DBClusterParameterGroupName")}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if tags:
        resource_translated["tags"] = tags
    return resource_translated



from typing import Any
from typing import Dict


def convert_raw_domain_identity_to_present(
    hub,
    ctx,
    raw_resource: str,
) -> Dict[str, Any]:
    """
    Util functions to convert raw resource state to present input format for SES Domain Identity.

    """

    resource_translated = {
        "name": raw_resource,
        "domain": raw_resource,
        "resource_id": raw_resource,
    }

    return resource_translated


def convert_raw_identity_notification_topic_to_present(
    hub,
    attributes: Dict[str, Any],
    identity: str,
    notification_type: str = None,
) -> Dict[str, Any]:
    """
    Util functions to convert raw resource state to present input format for SES identity notification topic.

    """

    resource_translated = {"identity": identity}
    if not notification_type:
        return None

    resource_translated["resource_id"] = identity + "::" + notification_type
    resource_translated["notification_type"] = notification_type
    resource_translated["sns_topic"] = attributes.get(f"{notification_type}Topic")
    resource_translated["enabled"] = attributes.get(
        f"HeadersIn{notification_type}NotificationsEnabled"
    )

    return resource_translated



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


async def convert_raw_base_path_mapping_to_present(
    hub,
    raw_resource: Dict[str, Any],
    domain_name: str,
) -> Dict[str, Any]:
    """
    Convert AWS API Gateway Resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        raw_resource(str): The AWS response to convert.
        domain_name(str): The domain name of the BasePathMapping resource to create.

    Returns:
        Dict[str, Any]
    """
    rest_api_id = raw_resource["restApiId"]
    base_path = raw_resource["basePath"]

    resource_id = f"{domain_name}/{base_path}"

    resource_parameters = OrderedDict(
        {
            "basePath": "base_path",
            "stage": "stage",
        }
    )

    resource_translated = {
        # "name": idem_resource_name,
        "domain_name": domain_name,
        "resource_id": resource_id,
        "rest_api_id": rest_api_id,
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def update(
    hub,
    ctx,
    domain_name: str,
    old_state: Dict[str, Any],
    updatable_parameters: Dict[str, Any],
):
    """
    Updates an API Gateway domain_name base path mapping.

    Args:
        domain_name(str):
            Apigateway the domain name for base path mapping
        old_state(dict):
            Previous state of the base path mapping resource.
        updatable_parameters(dict):
            Parameters from SLS File.

    Returns:
        Dict[str, Any]
    """
    result = dict(comment=(), result=True, ret=None)
    patch_operations = []
    new_state = copy.deepcopy(old_state)
    to_be_updated = False
    base_path = updatable_parameters["base_path"]
    f"{domain_name}/{base_path}"
    for key, value in updatable_parameters.items():
        if key == "rest_api_id":
            key_final = "restApiId"
        if key == "base_path":
            key_final = "basePath"
        if key == "stage":
            key_final = "stage"
        if (
            old_state.get(key) is not None
            and value is not None
            and value != old_state.get(key)
        ):
            patch_operations.append(
                {
                    "op": "replace",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            new_state[key] = value
            to_be_updated = True
    if ctx.get("test", False):
        if to_be_updated:
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.apigateway.base_path_mapping",
                name=old_state["domain_name"],
            )
            result["ret"] = new_state
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.apigateway.base_path_mapping",
                name=old_state["domain_name"],
            )
            return result
    if to_be_updated:
        update_ret = await hub.exec.boto3.client.apigateway.update_base_path_mapping(
            ctx,
            domainName=domain_name,
            basePath=base_path,
            patchOperations=patch_operations,
        )
        result["comment"] = hub.tool.aws.comment_utils.update_comment(
            resource_type="aws.apigateway.base_path_mapping",
            name=old_state["domain_name"],
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
    return result



"""Exec module for managing Amazon ECR repositories."""
from typing import Any
from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(
    hub, ctx, name: str, resource_id: str, registry_id: str = None
) -> Dict[str, Any]:
    """Retrieves an image repository in a registry.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str):
            The name of the ECR repository in Amazon Web Services.
        registry_id(str, Optional):
            The Amazon Web Services account ID associated with the registry that contains the repository
            to search. If you do not specify a registry, the default registry is assumed.

    Returns:
        Dict[str, Any]:
            Return image repository in a given registry.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.ecr.repository.get name='idem_name' resource_id='ecr_repository_name'

        Calling this exec module function from within a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ecr.repository.get
                - kwargs:
                    name: 'idem_name'
                    resource_id: 'ecr_repository_name'
    """
    result = dict(comment=[], ret=None, result=True)

    describe_ret = await hub.exec.boto3.client.ecr.describe_repositories(
        ctx=ctx, registryId=registry_id, repositoryNames=[resource_id]
    )

    # Case: Error
    if not describe_ret["result"]:
        # Do not return success=false when it is not found.
        if "RepositoryNotFoundException" in str(describe_ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ecr.repository", name=resource_id
                )
            )
            result["comment"] += list(describe_ret["comment"])
            return result

        result["comment"] += list(describe_ret["comment"])
        result["result"] = False
        return result

    # Case: Empty results
    if not describe_ret["ret"].get("repositories"):
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ecr.repository", name=resource_id
            )
        )
        return result

    # Case: More than one found
    if len(describe_ret["ret"].get("repositories")) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_type="aws.ecr.repository",
                resource_id=resource_id,
            )
        )

    # Case: One matching record is found (If more than one is found, then taking first)
    repository = describe_ret["ret"].get("repositories")[0]

    tags_results = await hub.tool.aws.ecr.tag.get_resource_tags(
        ctx=ctx, resource_arn=repository.get("repositoryArn")
    )

    result["ret"] = hub.tool.aws.ecr.conversion_utils.convert_raw_repository_to_present(
        raw_resource=repository,
        tags=tags_results.get("ret"),
        idem_resource_name=name,
    )

    return result


async def list_(
    hub, ctx, registry_id: str = None, repository_names: List[str] = None
) -> Dict[str, Any]:
    """Retrieves image repositories in a registry.

    Args:
        registry_id(str, Optional):
            The Amazon Web Services account ID associated with the registry that
            contains  the  repositories to be described. If you do not specify a
            registry, the default registry is assumed.
        repository_names(list[str], Optional):
            A list of repositories to describe.

    Returns:
        Dict[str, Any]:
            Return image repositories in a given registry.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.ecr.repository.list

        Calling this exec module function with repository name from the cli:

        .. code-block:: bash

            idem exec aws.ecr.repository.list repository_names=['repository_name_01', 'repository_name_02']

        Calling this exec module function from within a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ecr.repository.list
    """
    result = dict(comment=[], ret=[], result=True)

    ret = await hub.exec.boto3.client.ecr.describe_repositories(
        ctx=ctx, registryId=registry_id, repositoryNames=repository_names
    )

    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    if not ret["ret"].get("repositories"):
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ecr.repository", name=None
            )
        )
        return result

    for repository in ret["ret"]["repositories"]:
        tags_results = await hub.tool.aws.ecr.tag.get_resource_tags(
            ctx=ctx, resource_arn=repository.get("repositoryArn")
        )

        result["ret"].append(
            hub.tool.aws.ecr.conversion_utils.convert_raw_repository_to_present(
                raw_resource=repository,
                idem_resource_name=repository.get("repositoryName"),
                tags=tags_results.get("ret"),
            )
        )

    return result



"""Exec functions for AWS API Gateway v2 API resources."""
from typing import Any
from typing import Dict


__func_alias__ = {"list_": "list"}


async def get(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    """
    Get an API resource from AWS

    Args:
        name(str):
            The name of the Idem state

        resource_id(str):
            The API identifier

    Returns:
        Dict[str, Any]:
            Returns API resource

    Examples:
        Call from the CLI:

        .. code-block:: bash

            idem exec aws.apigatewayv2.api.get name="my_resource" resource_id="resource_id"

        Call from code:

        .. code-block:: python

            async def my_func(hub, ctx, name:str, resource_id:str):
                await hub.exec.aws.apigatewayv2.api.get(ctx, name, resource_id)

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.apigatewayv2.api.get
                - kwargs:
                    name: my_resource
                    resource_id: resource_id
    """
    result = dict(comment=[], ret=None, result=True)

    ret = await hub.exec.boto3.client.apigatewayv2.get_api(ctx, ApiId=resource_id)
    if not ret["result"]:
        if "NotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.apigatewayv2.api", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    result["ret"] = hub.tool.aws.apigatewayv2.api.convert_raw_api_to_present(
        raw_resource=ret["ret"]
    )
    return result


async def list_(hub, ctx) -> Dict[str, Any]:
    """
    Get a list of API resource from AWS

    Args:
        None

    Returns:
        Dict[str, Any]:
            Returns API resources

    Examples:
        Call from the CLI:

        .. code-block:: bash

            idem exec aws.apigatewayv2.api.list

        Call from code:

        .. code-block:: python

            async def my_func(hub, ctx):
                await hub.exec.aws.apigatewayv2.api.list(ctx)

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.apigatewayv2.api.list
    """
    result = dict(comment=[], ret=None, result=True)

    ret = await hub.exec.boto3.client.apigatewayv2.get_apis(ctx)

    if not ret["result"]:
        if "NotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.apigatewayv2.api"
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    result["ret"] = []
    for api in ret["ret"]["Items"]:
        result["ret"].append(
            hub.tool.aws.apigatewayv2.api.convert_raw_api_to_present(raw_resource=api)
        )
    return result



"""Conversion util functions to support AWS WAFV2 resources."""
import json
from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_web_acl_resource_association_to_present(
    hub,
    web_acl_arn: str,
    resource_arn: str,
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    """Util functions to convert raw resource state from Associate Web ACL to present input format.

    Args:
        web_acl_arn:
            The Amazon Resource Name (ARN) of the web ACL that you want to associate with the resource.
        resource_arn:
            The Amazon Resource Name (ARN) of the resource to associate with the web ACL.
        idem_resource_name:
            The idem name for the resource.

    Returns:
        Dict[str, Any]
    """
    resource_id = resource_arn
    translated_resource = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "web_acl_arn": web_acl_arn,
        "resource_arn": resource_arn,
    }
    return translated_resource


def convert_raw_regex_pattern_set_to_present(
    hub, ctx, raw_resource: Dict[str, Any], scope: str, tags: Dict[str, str]
) -> Dict[str, Any]:
    """Util functions to convert raw resource state from regex pattern state to present input format.

    Args:
        raw_resource(dict): Old state of resource or existing resource details.
        scope(str): Specifies whether this is for an Amazon CloudFront distribution or for a regional application.
        tags(dict): Dict in the format of {tag-key: tag-value}

    Returns:
        Dict[str, Any]
    """
    resource_parameters = OrderedDict(
        {
            "Name": "name",
            "Description": "description",
            "RegularExpressionList": "regular_expression_list",
            "Id": "resource_id",
            "ARN": "arn",
        }
    )
    resource_translated = {"scope": scope, "tags": tags}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = json.loads(
                json.dumps(raw_resource.get(parameter_raw))
            )

    return resource_translated


async def convert_raw_web_acl_to_present(
    hub,
    ctx,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    scope: str = None,
) -> Dict[str, Any]:
    """Util functions to convert raw resource state from Web ACL to present input format.

    Args:
        raw_resource:
            Old state of resource or existing resource details.
        idem_resource_name:
            Resource name.
        scope:
            Specifies whether this is for an Amazon CloudFront distribution or for a regional application.

    Returns:
        Dict[str, Any]
    """

    result = dict(comment=(), result=True, ret=None)
    resource_id = raw_resource.get("Id")
    web_acl_arn = raw_resource.get("ARN")
    resource_parameters = OrderedDict(
        {
            "DefaultAction": "default_action",
            "Description": "description",
            "Rules": "rules",
            "VisibilityConfig": "visibility_config",
            "CustomResponseBodies": "custom_response_bodies",
            "CaptchaConfig": "captcha_config",
        }
    )
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "scope": scope,
        "web_acl_arn": web_acl_arn,
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = json.loads(
                json.dumps(raw_resource.get(parameter_raw))
            )

    ret_tag = await hub.exec.boto3.client.wafv2.list_tags_for_resource(
        ctx, ResourceARN=web_acl_arn
    )
    result["result"] = ret_tag["result"]
    if not result["result"]:
        result["comment"] = result["comment"] + ret_tag["comment"]
        return result
    if (
        ret_tag["ret"]
        and ret_tag["ret"]["TagInfoForResource"]
        and ret_tag["ret"]["TagInfoForResource"].get("TagList")
    ):
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            ret_tag["ret"]["TagInfoForResource"].get("TagList")
        )
    result["ret"] = resource_translated
    return result


def convert_raw_ip_set_to_present(
    hub,
    scope: str,
    raw_resource: Dict[str, Any],
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    """Convert AWS returned data structure to correct idem wafv2 ip_set resource to present state.

     Args:
        hub:
            The redistributed pop central hub.
        ctx:
            A dict with the keys/values for the execution of the Idem run located in
            `hub.idem.RUNS[ctx['run_name']]`.
        scope (str) --
            Specifies whether this is for an Amazon CloudFront distribution or for a regional
            application. A regional application can be an Application Load Balancer (ALB), an Amazon API Gateway
            REST API, or an AppSync GraphQL API.
            To work with CloudFront, you must also specify the Region US East (N. Virginia) as follows:
            CLI - Specify the Region when you use the CloudFront scope: --scope=CLOUDFRONT --region=us-east-1 .
            API and SDKs - For all calls, use the Region endpoint us-east-1.
        raw_resource:
            The aws response to convert
        tags (List, Optional):
            The tags of the resource. Defaults to None.

    Returns:
        Dict[str, Any]
    """
    resource_translated = {
        "scope": scope,
    }

    resource_parameters_keys_case_mapping = OrderedDict(
        {
            "Name": "name",
            "Id": "resource_id",
            "Description": "description",
            "IPAddressVersion": "ip_address_version",
            "Addresses": "addresses",
            "Tags": "tags",
            "ARN": "arn",
            "LockToken": "lock_token",
        }
    )
    if (
        raw_resource
        and raw_resource.get("ret")
        and raw_resource.get("ret").get("IPSet")
    ):
        for (
            camel_case_key,
            snake_case_key,
        ) in resource_parameters_keys_case_mapping.items():
            if raw_resource["ret"]["IPSet"].get(camel_case_key):
                resource_translated[snake_case_key] = raw_resource["ret"]["IPSet"].get(
                    camel_case_key
                )
    if (
        raw_resource
        and raw_resource.get("ret")
        and raw_resource.get("ret").get("LockToken")
    ):
        resource_translated["lock_token"] = raw_resource.get("ret").get("LockToken")

    if tags or tags == {}:
        resource_translated["tags"] = tags
    return resource_translated



from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name,
    resource_id: str,
) -> Dict:
    """
    Use an un-managed vpc_endpoint as a data-source. Supply resource_id as a filter.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str):
            An AWS vpc_endpoint resource_id to identify the resource.

    Returns:
        Dict[bool, list, dict or None]:

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec aws.ec2.vpc_endpoint.get name="idem_name" resource_id="resource_id"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ec2.vpc_endpoint.get
                - kwargs:
                    name: my_resource
                    resource_id: resource_id
    """

    result = dict(comment=[], ret=None, result=True)

    resource_ret = await hub.exec.boto3.client.ec2.describe_vpc_endpoints(
        ctx, VpcEndpointIds=[resource_id]
    )

    if not resource_ret["result"]:
        if "InvalidVpcEndpointId.NotFound" in str(resource_ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.vpc_endpoint", name=name
                )
            )
            result["comment"] += list(resource_ret["comment"])
            return result
        result["comment"] += list(resource_ret["comment"])
        result["result"] = resource_ret["result"]
        return result

    if not resource_ret["ret"]["VpcEndpoints"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.vpc_endpoint", name=name
            )
        )
        return result

    result[
        "ret"
    ] = hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_endpoint_to_present(
        raw_resource=resource_ret["ret"]["VpcEndpoints"][0], idem_resource_name=name
    )

    return result


async def list_(hub, ctx, name: str = None, filters=None) -> Dict:
    """
    Use an un-managed VPC Endpoints as a data-source.

    Args:
        name (str, Optional):
            The name of the Idem state.

        filters(list, Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_vpc_endpoints

    Returns:
        Dict[bool, list, dict or None]:

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.ec2.vpc_endpoint.list name="idem_name"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ec2.vpc_endpoint.list
                - kwargs:
                    name: my_resource

    """
    result = dict(comment=[], ret=[], result=True)
    if filters:
        filters = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
            filters=filters
        )
    ret = await hub.exec.boto3.client.ec2.describe_vpc_endpoints(ctx, Filters=filters)
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["VpcEndpoints"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.vpc_endpoint", name=name
            )
        )
        return result
    for vpc_endpoint in ret["ret"]["VpcEndpoints"]:
        resource_id = vpc_endpoint.get("VpcEndpointId")
        converted_resource = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_endpoint_to_present(
                vpc_endpoint, idem_resource_name=resource_id
            )
        )
        result["ret"].append(converted_resource)
    return result



"""State module for managing Amazon GuardDuty Detector."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    enable: bool = True,
    resource_id: str = None,
    client_token: str = None,
    finding_publishing_frequency: str = None,
    data_sources: make_dataclass(
        """Describes which data sources will be enabled for the detector."""
        "DataSourceConfiguration",
        [
            (
                "S3Logs",
                make_dataclass(
                    """Describes whether S3 data event logs are enabled as a data source."""
                    "S3LogsConfiguration",
                    [("Enable", bool)],
                ),
                field(default=None),
            ),
            (
                "Kubernetes",
                make_dataclass(
                    """Describes whether any Kubernetes logs are enabled as data sources."""
                    "KubernetesConfiguration",
                    [
                        (
                            "AuditLogs",
                            make_dataclass(
                                """The status of Kubernetes audit logs as a data source."""
                                "KubernetesAuditLogsConfiguration",
                                [("Enable", bool)],
                            ),
                        )
                    ],
                ),
                field(default=None),
            ),
            (
                "MalwareProtection",
                make_dataclass(
                    """Describes whether Malware Protection is enabled as a data source."""
                    "MalwareProtectionConfiguration",
                    [
                        (
                            "ScanEc2InstanceWithFindings",
                            make_dataclass(
                                """Describes the configuration of Malware Protection for EC2 instances with findings."""
                                "ScanEc2InstanceWithFindingsConfiguration",
                                [("EbsVolumes", bool, field(default=None))],
                            ),
                            field(default=None),
                        )
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    tags: Dict[str, str] = None,
) -> Dict[str, Any]:
    """Creates an Amazon GuardDuty detector.

    A detector is a resource that represents the GuardDuty service.
    To start using GuardDuty, you must create a detector in each Region where you enable the service.
    You can have only one detector per account per Region. All data sources are enabled in a new detector by default.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str, Optional):
            The ID of the detector in Amazon Web Services.
        enable(bool, Optional):
            A Boolean value that specifies whether the detector is to be enabled. Default value is ``True``.
        client_token(str, Optional):
            The idempotency token for the create request. This field is auto_populated if not provided.
        finding_publishing_frequency(str, Optional):
            A value that specifies how frequently updated findings are exported.
            Valid values are ``FIFTEEN_MINUTES``, ``ONE_HOUR``, ``SIX_HOURS``.
        data_sources(dict, Optional):
            Describes which data sources will be enabled for the detector.

            * S3Logs (*dict, Optional*):
                Describes whether S3 data event logs are enabled as a data source.

                * Enable (*bool*): The status of S3 data event logs as a data source.

            * Kubernetes (*dict, Optional*):
                Describes whether any Kubernetes logs are enabled as data sources.

                * AuditLogs (*dict*):
                    The status of Kubernetes audit logs as a data source.

                    * Enable (*bool*):
                        The status of Kubernetes audit logs as a data source.

            * MalwareProtection (*dict, Optional*):
                Describes whether Malware Protection is enabled as a data source.

                * ScanEc2InstanceWithFindings (*dict, Optional*):
                    Describes the configuration of Malware Protection for EC2 instances with findings.

                    EbsVolumes (*bool, Optional*):
                        Describes the configuration for scanning EBS volumes as data source.

        tags(dict, Optional):
            Dict in the format of ``{tag-key: tag-value}`` to associate with the detector.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_guardduty_detector]:
          aws.guaardduty.detector.present:
            - name: 'string'
            - enable: True|False
            - client_token: 'string'
            - finding_publishing_frequency: 'FIFTEEN_MINUTES|ONE_HOUR|SIX_HOURS'
            - data_sources:
                S3Logs:
                  Enable: True|False
                Kubernetes:
                  AuditLogs:
                    Enable: True|False
                MalwareProtection:
                  ScanEc2InstanceWithFindings:
                    EbsVolumes: True|False
            - tags:
                'string': 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_guardduty_detector:
              aws.guardduty.detector.present:
                - name: idem_test_guardduty_detector
                - enable: True
                - finding_publishing_frequency: 'ONE_HOUR'
                - data_sources:
                    S3Logs:
                      Enable: true
                - tags:
                    provider: idem

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.guardduty.detector.get(
            ctx, resource_id=resource_id, name=name
        )
        if not before["result"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
    if not before or not before["ret"]:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "enable": enable,
                    "finding_publishing_frequency": finding_publishing_frequency,
                    "data_sources": data_sources,
                    "tags": tags,
                },
            )
            result["comment"] = (f"Would create aws.guardduty.detector {name}",)
            return result
        try:
            ret = await hub.exec.boto3.client.guardduty.create_detector(
                ctx,
                Enable=enable,
                ClientToken=client_token,
                FindingPublishingFrequency=finding_publishing_frequency,
                DataSources=data_sources,
                Tags=tags,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = result["comment"] + (f"Created '{name}'",)
            resource_id = ret["ret"]["DetectorId"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
            return result

    else:
        try:
            account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
            account_id = account_details["ret"]["Account"]
            region_name = ctx["acct"]["region_name"]
            detector_arn = hub.tool.aws.arn_utils.build(
                service="guardduty",
                region=region_name,
                account_id=account_id,
                resource="detector/" + resource_id,
            )
            result["old_state"] = copy.deepcopy(before["ret"])
            plan_state = copy.deepcopy(result["old_state"])
            update_ret = await hub.tool.aws.guardduty.detector.update(
                ctx,
                before=before,
                detector_id=resource_id,
                finding_publishing_frequency=finding_publishing_frequency,
                data_sources=data_sources,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = update_ret["result"]
            resource_updated = resource_updated or bool(update_ret["ret"])
            if update_ret["ret"] and ctx.get("test", False):
                if "finding_publishing_frequency" in update_ret["ret"]:
                    plan_state["finding_publishing_frequency"] = update_ret["ret"][
                        "finding_publishing_frequency"
                    ]
                if "data_sources" in update_ret["ret"]:
                    plan_state["data_sources"] = update_ret["ret"]["data_sources"]

            if (tags is not None) and (result["old_state"].get("tags", {}) != tags):
                update_tags_ret = await hub.tool.aws.guardduty.detector.update_tags(
                    ctx=ctx,
                    resource_arn=detector_arn,
                    old_tags=result["old_state"].get("tags", {}),
                    new_tags=tags,
                )
                if not result["result"]:
                    result["comment"] = update_tags_ret["comment"]
                    result["result"] = False
                    return result
                result["comment"] = result["comment"] + update_tags_ret["comment"]
                result["result"] = result["result"] and update_tags_ret["result"]
                resource_updated = resource_updated or bool(update_tags_ret["ret"])
                if ctx.get("test", False) and update_tags_ret["ret"] is not None:
                    plan_state["tags"] = update_tags_ret["ret"].get("tags")
            if not resource_updated:
                result["comment"] = result["comment"] + (f"{name} already exists",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
            return result

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or (not before["ret"]) or resource_updated:
            after = await hub.exec.aws.guardduty.detector.get(
                ctx, resource_id=resource_id, name=name
            )
            if not after["result"] or not after["ret"]:
                result["result"] = False
                result["comment"] = after["comment"]
                return result
            result["new_state"] = copy.deepcopy(after["ret"])
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str,
) -> Dict[str, Any]:
    """Deletes an Amazon GuardDuty detector.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str):
            The ID of the detector in Amazon Web Services.

    Request syntax:
      .. code-block:: sls

        [idem_test_aws_guardduty_detector]:
          aws.guardduty.detector.absent:
            - name: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_guardduty_detector:
              aws.guardduty.detector.absent:
                - name: idem_test_guardduty_detector
                - resource_id: cebf7ced6562d943d61f76a915e32563
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.aws.guardduty.detector.get(
        ctx, resource_id=resource_id, name=name
    )
    if not before["result"] or not before["ret"]:
        result["comment"] = (f"'{name}' already absent",)
    elif ctx.get("test", False):
        result["old_state"] = copy.deepcopy(before["ret"])
        result["comment"] = (f"Would delete aws.guardduty.detector",)
        return result
    else:
        result["old_state"] = copy.deepcopy(before["ret"])
        try:
            ret = await hub.exec.boto3.client.guardduty.delete_detector(
                ctx, DetectorId=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = (f"Deleted '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes AWS GuardDuty detectors in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.guardduty.detector
    """
    result = {}
    ret = await hub.exec.boto3.client.guardduty.list_detectors(ctx)

    if not ret["ret"]["DetectorIds"]:
        hub.log.debug(f"Could not list detector {ret['comment']}")
        return {}

    for resource_id in ret["ret"]["DetectorIds"]:
        detector_id = resource_id
        detector = await hub.exec.boto3.client.guardduty.get_detector(
            ctx, DetectorId=resource_id
        )

        resource_translated = await hub.tool.aws.guardduty.conversion_utils.convert_raw_detector_to_present(
            ctx, raw_resource=detector, idem_resource_name=detector_id
        )

        result[detector_id] = {
            "aws.guardduty.detector.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_stream_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    tags: Dict[str, str] = None,
) -> Dict[str, Any]:
    """
    Util function to convert raw resource state from AWS Kinesis Stream to present input format.

    Args:
        raw_resource(Dict[str, Any]):
            The AWS response from 'Kinesis DescribeStreamSummary' to convert.
        idem_resource_name(str, Optional): An Idem name of the resource.
        tags(Dict[str, str], Optional): The AWS Kinesis Stream tags.

    Returns:
        Dict[str, Any]
    """

    resource_parameters = OrderedDict(
        {
            "OpenShardCount": "shard_count",
            "RetentionPeriodHours": "retention_period_hours",
            "EncryptionType": "encryption_type",
            "KeyId": "key_id",
            "StreamARN": "stream_arn",
        }
    )
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": idem_resource_name,
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if (
        raw_resource.get("EnhancedMonitoring")
        and len(raw_resource["EnhancedMonitoring"])
        and len(raw_resource["EnhancedMonitoring"][0].get("ShardLevelMetrics")) > 0
    ):
        resource_translated.update(
            {
                "shard_level_metrics": (
                    (raw_resource.get("EnhancedMonitoring")[0]).get("ShardLevelMetrics")
                ).copy()
            }
        )
    if raw_resource.get("StreamModeDetails"):
        resource_translated.update(
            {"stream_mode_details": (raw_resource.get("StreamModeDetails")).copy()}
        )
    if tags:
        resource_translated["tags"] = tags.copy()
    return resource_translated



"""State module for managing Cloudfront Origin request policies."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    headers_config: make_dataclass(
        "OriginRequestPolicyHeadersConfig",
        [
            ("HeaderBehavior", str),
            (
                "Headers",
                make_dataclass(
                    "Headers",
                    [("Quantity", int), ("Items", List[str], field(default=None))],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    cookies_config: make_dataclass(
        "OriginRequestPolicyCookiesConfig",
        [
            ("CookieBehavior", str),
            (
                "Cookies",
                make_dataclass(
                    "CookieNames",
                    [("Quantity", int), ("Items", List[str], field(default=None))],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    query_strings_config: make_dataclass(
        "OriginRequestPolicyQueryStringsConfig",
        [
            ("QueryStringBehavior", str),
            (
                "QueryStrings",
                make_dataclass(
                    "QueryStringNames",
                    [("Quantity", int), ("Items", List[str], field(default=None))],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    comment: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Creates an origin request policy.

    After you create an origin request policy, you can attach it to one or morecache behaviors. When it’s attached to a
    cache behavior, the origin request policy determines the values th includes the following: The request body and the
    URL path (without the domain name) from the viewer request.

    The headers that CloudFront automatically includes in every origin request, including Host, User-Agent, and
    X-Amz-Cf-Id.   All HTTP headers, cookies, and URL query strings that are specified in the cache policy or the
    origin request policy. These can include items from the viewer request and, in the case of headers, additional
    ones that are added by CloudFront.   CloudFront sends a request when it can’t find a valid object in its cache
    that matches the request. If you want to send values to the origin and also include them in the cache key, use
    CachePolicy. For more information about origin request policies, see Controlling origin requests in the Amazon
    CloudFront Developer Guide.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        comment (str, Optional):
            A comment to describe the origin request policy. The comment cannot be longer than 128
            characters.

        headers_config (dict):
            The HTTP headers to include in origin requests. These can include headers from viewer requests
            and additional headers added by CloudFront.

            * HeaderBehavior (str):
                Determines whether any HTTP headers are included in requests that CloudFront sends to the
                origin. Valid values are:    none – HTTP headers are not included in requests that CloudFront
                sends to the origin. Even when this field is set to none, any headers that are listed in a
                CachePolicy are included in origin requests.    whitelist – The HTTP headers that are listed in
                the Headers type are included in requests that CloudFront sends to the origin.    allViewer –
                All HTTP headers in viewer requests are included in requests that CloudFront sends to the
                origin.    allViewerAndWhitelistCloudFront – All HTTP headers in viewer requests and the
                additional CloudFront headers that are listed in the Headers type are included in requests that
                CloudFront sends to the origin. The additional headers are added by CloudFront.

                * Headers (Dict[str, Any], Optional):
                    Contains a list of HTTP header names.
                    * Quantity (int): The number of header names in the Items list.
                    * Items (list[str], Optional): A list of HTTP header names.

        cookies_config (dict):
            The cookies from viewer requests to include in origin requests.

            * CookieBehavior (str):
                Determines whether cookies in viewer requests are included in requests that CloudFront sends to
                the origin. Valid values are:    none – Cookies in viewer requests are not included in requests
                that CloudFront sends to the origin. Even when this field is set to none, any cookies that are
                listed in a CachePolicy are included in origin requests.    whitelist – The cookies in viewer
                requests that are listed in the CookieNames type are included in requests that CloudFront sends
                to the origin.    all – All cookies in viewer requests are included in requests that CloudFront
                sends to the origin.

                * Cookies (Dict[str, Any], Optional):
                    Contains a list of cookie names.
                    * Quantity (int): The number of cookie names in the Items list.
                    * Items (list[str], Optional): A list of cookie names.

        query_strings_config (dict):
            The URL query strings from viewer requests to include in origin requests.

            * QueryStringBehavior (str):
                Determines whether any URL query strings in viewer requests are included in requests that
                CloudFront sends to the origin. Valid values are:    none – Query strings in viewer requests are
                not included in requests that CloudFront sends to the origin. Even when this field is set to
                none, any query strings that are listed in a CachePolicy are included in origin requests.
                whitelist – The query strings in viewer requests that are listed in the QueryStringNames type
                are included in requests that CloudFront sends to the origin.    all – All query strings in
                viewer requests are included in requests that CloudFront sends to the origin.

                * QueryStrings (Dict[str, Any], Optional):
                    Contains a list of the query strings in viewer requests that are included in requests that CloudFront sends to the origin.
                    * Quantity (int): The number of query string names in the Items list.
                    * Items (list[str], Optional): A list of query string names.

    Request Syntax:

        .. code-block:: sls

            resource_is_present:
              aws.cloudfront.origin_request_policy.present:
                - name: "string"
                - comment: "string"
                - headers_config:
                    HeaderBehavior: "dict"
                - cookies_config:
                    CookieBehavior: "dict"
                - query_strings_config:
                    QueryStringBehavior: "dict"

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.cloudfront.origin_request_policy.present:
                - name: value
                - comment: value
                - headers_config:
                    HeaderBehavior: none
                - cookies_config:
                    CookieBehavior: none
                - query_strings_config:
                    QueryStringBehavior: none
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.cloudfront.origin_request_policy.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])

        # check update
        update_ret = (
            await hub.tool.aws.cloudfront.cloudfront_utils.update_origin_request_policy(
                ctx=ctx,
                name=name,
                before=result["old_state"],
                resource_id=resource_id,
                comment=comment,
                headers_config=headers_config,
                cookies_config=cookies_config,
                query_strings_config=query_strings_config,
                if_match=before["ret"].get("etag"),
            )
        )
        result["comment"] = update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = bool(update_ret["ret"])
        if update_ret["ret"]:
            if ctx.get("test", False):
                for modified_param in update_ret["ret"]:
                    plan_state[modified_param] = update_ret["ret"][modified_param]
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.cloudfront.origin_request_policy", name=name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.cloudfront.origin_request_policy", name=name
                )

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "comment": comment,
                    "headers_config": headers_config,
                    "cookies_config": cookies_config,
                    "query_strings_config": query_strings_config,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.cloudfront.origin_request_policy", name=name
            )
            return result

        origin_request_policy_config = {
            "Comment": comment,
            "Name": name,
            "HeadersConfig": headers_config,
            "CookiesConfig": cookies_config,
            "QueryStringsConfig": query_strings_config,
        }
        origin_request_policy_config = (
            hub.tool.aws.cloudfront.cloudfront_utils.sanitize_policy_config(
                origin_request_policy_config
            )
        )
        ret = await hub.exec.boto3.client.cloudfront.create_origin_request_policy(
            ctx, OriginRequestPolicyConfig=origin_request_policy_config
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.cloudfront.origin_request_policy", name=name
        )
        resource_id = ret["ret"]["OriginRequestPolicy"]["Id"]

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not before) or resource_updated:
        after = await hub.exec.aws.cloudfront.origin_request_policy.get(
            ctx, name=name, resource_id=resource_id
        )
        if not after["result"]:
            result["result"] = False
            result["comment"] = after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes an origin request policy.

    You cannot delete an origin request policy if it’s attached to any cache behaviors. First update your distributions
    to remove the origin request policy from all cache behaviors, then delete the origin request policy. To delete an
    origin request policy, you must provide the policy’s identifier and version.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str):
            An identifier of the resource in the provider.

    Request Syntax:

        .. code-block:: sls

            resource_is_absent:
              aws_auto.cloudfront.origin_request_policy.absent:
                - name: "string"
                - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws_auto.cloudfront.origin_request_policy.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.cloudfront.origin_request_policy", name=name
        )
        return result

    before = await hub.exec.aws.cloudfront.origin_request_policy.get(
        ctx, name=name, resource_id=resource_id
    )

    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.cloudfront.origin_request_policy", name=name
        )

    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.cloudfront.origin_request_policy", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]

        ret = await hub.exec.boto3.client.cloudfront.delete_origin_request_policy(
            ctx, Id=resource_id, IfMatch=before["ret"]["etag"]
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.cloudfront.origin_request_policy", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws_auto.cloudfront.origin_request_policy
    """

    result = {}

    ret = await hub.exec.aws.cloudfront.origin_request_policy.list(ctx)
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    for policy in ret["ret"]:
        resource_id = policy.get("resource_id")

        result[resource_id] = {
            "aws.cloudfront.origin_request_policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in policy.items()
            ]
        }

    return result



"""Exec functions for S3 bucket's logging."""
from typing import Dict


async def get(hub, ctx, name: str, resource_id: str) -> Dict:
    """Returns the logging configuration for the S3 bucket.

    Args:
        name(str):
            The name of the Idem state.
        resource_id(str):
            AWS S3 bucket name.

    Examples:
        Calling from the CLI:

        .. code-block:: bash

            $ idem exec aws.s3.bucket_logging.get name="bucket-logging-name" resource_id="bucket-logging-id"

        Using in a state:

        .. code-block:: yaml

            get_a_bucket_logging:
              exec.run:
                - path: aws.s3.bucket_logging.get
                - kwargs:
                   name: bucket-logging-name
                   resource_id: bucket-logging-id
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.s3.get_bucket_logging(ctx, Bucket=resource_id)

    if not ret["result"]:
        if "NoSuchBucket" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.s3.bucket_logging", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    if "LoggingEnabled" not in ret["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.s3.bucket_logging", name=name
            )
        )
        result["comment"] += list(ret["comment"])
        return result

    result[
        "ret"
    ] = hub.tool.aws.s3.conversion_utils.convert_raw_bucket_logging_to_present(
        ctx=ctx, raw_resource=ret["ret"], bucket_name=resource_id
    )
    return result



"""Exec module for AWS API Gateway Base Path Mapping resources."""
__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    domain_name: str = None,
    base_path: str = None,
    resource_id: str = None,
):
    """
    Get a single API Gateway base path mapping from AWS. The function returns None if no resource is found.
    The get api call requires either of resource_id or the combination of base_path and domain name. So the user is free to choose between either the resource_id or base_path and domain name. But, both these sets are not required.
    When using domain_name and base_path as input parameters, both are mandatory. While, using resource_id, the other two parameters can be skipped, and hence they come as optional parameters in the signature.

    Args:
        domain_name(str, Optional):
            The parent domain name of the base_path_mapping resource to be described.
        base_path(str, Optional):
            The base path name that callers of the API must provide as part
            of the URL after the domain name. This value must be unique for all of the mappings
            across a single API. Specify '(none)' if you do not want callers to specify any base
            path name after the domain name.
        resource_id(str, Optional):
            Idem Resource ID that is generated once the resource is created.


    Examples:
        Calling from the CLI:

        .. code-block:: bash

            $ idem exec aws.apigateway.base_path_mapping.get domain_name="unmanaged_apigateway_base_path_mapping_domain_name" base_path = "unmanaged_apigateway_base_path_mapping"

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.apigateway.base_path_mapping.get
                - kwargs:
                    domain_name: domain_name_of_base_path
                    base_path: value

        .. code-block:: bash

            $ idem exec aws.apigateway.base_path_mapping.get resource_id = "base_path_mapping_unique_id"

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.apigateway.base_path_mapping.get
                - kwargs:
                    resource_id: unique_id_created
    """
    result = dict(comment=[], ret=None, result=True)
    # If resource_id is specified then it overrides any domain_name and base_path provided
    if resource_id:
        domain_name = resource_id.split("/")[0]
        base_path = resource_id.split("/")[1]
    get_base_path_mapping = (
        await hub.exec.boto3.client.apigateway.get_base_path_mapping(
            ctx, domainName=domain_name, basePath=base_path
        )
    )
    if not get_base_path_mapping["result"]:
        if "NotFoundException" in str(get_base_path_mapping["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.apigateway.base_path_mapping", name=domain_name
                )
            )
            result["comment"] += list(get_base_path_mapping["comment"])
            return result
        result["comment"] += list(get_base_path_mapping["comment"])
        result["result"] = False
        return result

    result[
        "ret"
    ] = await hub.tool.aws.apigateway.base_path_mapping.convert_raw_base_path_mapping_to_present(
        raw_resource=get_base_path_mapping["ret"],
        domain_name=domain_name,
    )
    return result


async def list_(hub, ctx, domain_name: str):
    """
    Get the list of base path mappings for the given domain name for AWS APIGateway.

    Args:
        domain_name(str):
            The parent domain name of the base_path_mapping resource to be described.

    Returns:
        Dict[str, Any]

    Examples:
        Calling from the CLI:

        .. code-block:: bash

            idem exec aws.apigateway.base_path_mapping.list domain_name="unmanaged_apigateway_base_path_mapping_domain_name"

        Calling this exec module function from within a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.apigateway.base_path_mapping.list
                - kwargs:
                    domain_name: domain_name_of_base_paths
    """
    result = dict(comment=[], ret=[], result=True)
    get_base_path_mappings = (
        await hub.exec.boto3.client.apigateway.get_base_path_mappings(
            ctx, domainName=domain_name
        )
    )
    if not get_base_path_mappings["result"]:
        result["comment"] += list(get_base_path_mappings["comment"])
        result["result"] = False
        return result
    if (
        "NotFoundException" in str(get_base_path_mappings["comment"])
        or get_base_path_mappings["ret"] == {}
        or get_base_path_mappings["ret"]["items"] == []
    ):
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.apigateway.base_path_mapping", name="domain_name"
            )
        )
        return result

    for base_path_map in get_base_path_mappings["ret"]["items"]:
        resource_translated = await hub.tool.aws.apigateway.base_path_mapping.convert_raw_base_path_mapping_to_present(
            raw_resource=base_path_map,
            domain_name=domain_name,
        )
        result["ret"].append(resource_translated)
    return result



"""State module for managing Amazon API Gateway Deployments."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    rest_api_id: str,
    resource_id: str = None,
    stage_name: str = None,
    stage_description: str = None,
    description: str = None,
    cache_cluster_enabled: bool = None,
    cache_cluster_size: str = None,
    variables: Dict = None,
    canary_settings: make_dataclass(
        "canarySettings",
        [
            ("percentTraffic", str),
            ("stageVariableOverrides", Dict, field(default=None)),
        ],
    ) = None,
    tracing_enabled: bool = None,
) -> Dict[str, Any]:
    """Creates an API Gateway Deployment resource.

    Args:
        name(str):
            An Idem name of the resource.

        rest_api_id(str):
            The string identifier of the associated RestApi.

        resource_id(str, Optional):
            AWS API Gateway Deployment ID.

        stage_name(str, Optional):
            The name of the Stage resource for the Deployment resource to create. This parameter is only valid during
            creation. Defaults to None.

        stage_description(str, Optional):
            The description of the Stage resource for the Deployment resource to create. This parameter is only valid
            during creation. Defaults to None.

        description(str, Optional):
            The description for the Deployment resource to create. Defaults to None.

        cache_cluster_enabled(bool, Optional):
            Enables a cache cluster for the Stage resource specified in the input. This parameter is only valid during
            creation. Defaults to None.

        cache_cluster_size(str, Optional):
            Specifies the cache cluster size for the Stage resource specified in the input, if a cache cluster is enabled.
            This parameter is only valid during creation. Defaults to None.

        variables(dict, Optional):
            A map that defines the stage variables for the Stage resource that is associated with the new deployment.
            Variable names can have alphanumeric and underscore characters, and the values must match [A-Za-z0-9-._~:/?#&=,]+.
            This parameter is only valid during creation. Defaults to None.

        canary_settings(dict, Optional):
            The input configuration for the canary deployment when the deployment is a canary release deployment. This
            parameter is only valid during creation. Defaults to None.

        tracing_enabled(bool, Optional):
            Specifies whether active tracing with X-ray is enabled for the Stage. This parameter is only valid during
            creation. Defaults to None.

    Request Syntax:
        .. code-block:: sls

            [idem_test_aws_apigateway_deployment]:
            aws.apigateway.deployment.present:
                - name: 'string'
                - resource_id: 'string'
                - stage_name: 'string'
                - stage_description: 'string'
                - description: 'string'
                - cache_cluster_enabled: bool
                - cache_cluster_size: 'string'
                - variables:
                - 'string' : 'string'
                - canary_settings:
                - percentTraffic: float
                - stageVariableOverrides:
                    - 'string: 'string'
                useStageCache: bool
                - tracing_enabled: bool


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.apigateway.deployment.present:
                - name: value
                - rest_api_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.apigateway.deployment.get(
            ctx,
            rest_api_id=rest_api_id,
            resource_id=resource_id,
            name=name,
        )
        if not before["result"] or not before["ret"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

        result["old_state"] = before["ret"]
        result["new_state"] = copy.deepcopy(result["old_state"])

        updatable_parameters = {
            "description": description,
        }

        update_deployment_ret = await hub.tool.aws.apigateway.deployment.update(
            ctx,
            old_state=result["old_state"],
            updatable_parameters=updatable_parameters,
        )

        result["comment"] = result["comment"] + update_deployment_ret["comment"]

        if not update_deployment_ret["result"]:
            result["result"] = False
            return result
        resource_updated = bool(update_deployment_ret["ret"])

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": resource_id,
                    "description": description,
                    "rest_api_id": rest_api_id,
                    "stage_name": stage_name,
                    "stage_description": stage_description,
                    "cache_cluster_enabled": cache_cluster_enabled,
                    "cache_cluster_size": cache_cluster_size,
                    "variables": variables,
                    "canary_settings": canary_settings,
                    "tracing_enabled": tracing_enabled,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigateway.deployment", name=name
            )
            return result

        ret = await hub.exec.boto3.client.apigateway.create_deployment(
            ctx,
            restApiId=rest_api_id,
            stageName=stage_name,
            stageDescription=stage_description,
            description=description,
            cacheClusterEnabled=cache_cluster_enabled,
            cacheCluserSize=cache_cluster_size,
            variables=variables,
            canarySettings=canary_settings,
            tracingEnabled=tracing_enabled,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result

        resource_id = ret["ret"]["id"]
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigateway.deployment", name=name
        )

    if (not before) or resource_updated:
        after = await hub.exec.aws.apigateway.deployment.get(
            ctx,
            name=name,
            resource_id=resource_id,
            rest_api_id=rest_api_id,
        )
        if not (after["result"] and after["ret"]):
            result["result"] = False
            result["comment"] = result["comment"] + after["comment"]
            return result

        result["new_state"] = after["ret"]
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    rest_api_id: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Deletes a Deployment resource.

    Deleting a deployment will only succeed if there are no Stage resources associated with it.

    Args:
        name(str):
            An Idem name of the resource.

        rest_api_id(str):
            The string identifier of the associated RestApi.

        resource_id(str, Optional):
            AWS API Gateway Deployment ID. Idem automatically considers this resource being absent if this field is not
            specified.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.apigateway.deployment.absent:
                - name: value
                - resource_id: value
                - rest_api_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        # if resource_id isn't specified, the resource is considered to be absent.
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.deployment", name=name
        )
        return result

    before_ret = await hub.exec.aws.apigateway.deployment.get(
        ctx,
        name=name,
        resource_id=resource_id,
        rest_api_id=rest_api_id,
    )

    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result

    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.deployment", name=name
        )
        return result

    elif ctx.get("test", False):
        result["old_state"] = before_ret["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigateway.deployment", name=name
        )
        return result

    else:
        result["old_state"] = before_ret["ret"]
        delete_ret = await hub.exec.boto3.client.apigateway.delete_deployment(
            ctx, restApiId=rest_api_id, deploymentId=resource_id
        )

        if not delete_ret["result"]:
            result["result"] = False
            result["comment"] = delete_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.apigateway.deployment", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the API Gateway Deployments.

    Returns a list of apigateway.deployment descriptions

    Returns:
        Dict[str, Any]


    Examples:
        .. code-block:: bash

            $ idem describe aws.apigateway.deployment

    """
    result = {}

    get_rest_apis_ret = await hub.exec.boto3.client.apigateway.get_rest_apis(ctx)
    if not get_rest_apis_ret["result"]:
        hub.log.debug(f"Could not get Rest Apis {get_rest_apis_ret['comment']}")
        return result

    if get_rest_apis_ret["ret"]["items"] is not None:
        for rest_api in get_rest_apis_ret["ret"]["items"]:
            rest_api_id = rest_api.get("id")
            get_deployments_ret = (
                await hub.exec.boto3.client.apigateway.get_deployments(
                    ctx, restApiId=rest_api_id
                )
            )
            if not get_deployments_ret["result"]:
                hub.log.debug(
                    f"Could not describe deployments for Rest API {get_deployments_ret['comment']}. Skipping this resource "
                    f"and continuing."
                )
                continue

            for deployment in get_deployments_ret["ret"]["items"]:
                resource_id = deployment.get("id")
                resource_translated = await hub.tool.aws.apigateway.deployment.convert_raw_deployment_to_present(
                    raw_resource=deployment,
                    resource_id=resource_id,
                    rest_api_id=rest_api_id,
                )
                result[resource_translated["resource_id"]] = {
                    "aws.apigateway.deployment.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in resource_translated.items()
                    ]
                }

    return result



"""Exec functions for backup vault."""
from typing import Dict


__func_alias__ = {"list_": "list"}


async def get(hub, ctx, name: str, resource_id: str = None):
    """Returns the backup vault.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str):
            AWS Backup vault name.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.backup.backup_vault.get name="idem_name"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.backup.backup_vault.get(
                    ctx, name=name, resource_id=resource_id
                )
    """
    result = dict(comment=[], ret=None, result=True)

    if not resource_id:
        resource_id = name
    ret = await hub.exec.boto3.client.backup.describe_backup_vault(
        ctx, BackupVaultName=resource_id
    )
    if not ret["result"]:
        if "ResourceNotFoundException" or "ClientError:" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.backup.backup_vault", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["result"] = False
        result["comment"] += list(ret["comment"])
        return result

    if not ret["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.backup.backup_vault", name=name
            )
        )
        result["comment"] += list(ret["comment"])
        return result

    arn = ret["ret"]["BackupVaultArn"]
    tags_ret = await hub.exec.boto3.client.backup.list_tags(ctx, ResourceArn=arn)
    if not tags_ret["result"]:
        result["result"] = False
        result["comment"] = tags_ret["comment"]
        return result

    if tags_ret["ret"].get("Tags"):
        ret["ret"]["Tags"] = tags_ret["ret"]["Tags"]

    result[
        "ret"
    ] = hub.tool.aws.backup.conversion_utils.convert_raw_backup_vault_to_present(
        ctx=ctx,
        raw_resource=ret["ret"],
        vault_name=resource_id,
    )

    return result


async def list_(hub, ctx, name: str = None) -> Dict:
    """Fetch a list of backup vaults from AWS.

    The function returns empty list when no resource is found.

    Args:
        name (str, Optional):
            The name of the Idem state for logging.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The AWS backup vault list in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.backup.backup_vault.list name="idem_name"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.backup.backup_vault.list(ctx, name)

    """
    result = dict(comment=[], ret=[], result=True)

    ret = await hub.exec.boto3.client.backup.list_backup_vaults(ctx)

    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["BackupVaultList"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.backup.backup_vault", name=name
            )
        )
        return result

    for backup_vault in ret["ret"]["BackupVaultList"]:
        tags_ret = await hub.exec.boto3.client.backup.list_tags(
            ctx, ResourceArn=backup_vault.get("BackupVaultArn")
        )

        if not tags_ret["result"]:
            result["result"] = False
            result["comment"] = tags_ret["comment"]
            return result

        if tags_ret["ret"].get("Tags"):
            backup_vault["Tags"] = tags_ret["ret"]["Tags"]

        result["ret"].append(
            hub.tool.aws.backup.conversion_utils.convert_raw_backup_vault_to_present(
                ctx=ctx,
                raw_resource=backup_vault,
                vault_name=backup_vault["BackupVaultName"],
            )
        )
    return result



"""Exec module for managing launch template."""
from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name,
    resource_id: str = None,
    filters: List = None,
) -> Dict:
    """Get a Launch template resource from AWS. Supply one of the inputs as the filter.

    Args:
        name(str):
            The name of the Idem state.
        resource_id(string, Optional):
            AWS Launch template id to identify the resource.
        filters(list, Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_launch_templates

    Returns:
        Dict[str, Any]:
            Returns launch_template in present format

    Examples:
        Calling this exec module function from the cli with filters

        .. code-block:: bash

            idem exec aws.ec2.launch_template.get  name=my_resource filters=[{'name': 'tag:Name', 'values': ['value']}]

        Calling this exec module function from within a state module in pure python.

        .. code-block:: python

            async def state_function(hub, ctx, name, service_namespace, scaling_resource_id, scalable_dimension):
                await hub.exec.aws.ec2.launch_template.get(
                    ctx=ctx,
                    name=name,
                    resource_id=resource_id,
                    filters=filters
                )

        Using in a state:

        .. code-block:: yaml

           my_unmanaged_resource:
              exec.run:
                - path: aws.ec2.launch_template.get
                - kwargs:
                    name: my_resource
                    filters:
                      - name: 'launch-template-name'
                        values: [ "template-name" ]
                      - name: 'tag:Key'
                        values: [ "value" ]
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.ec2.launch_template.search_raw(
        ctx=ctx,
        name=name,
        resource_id=resource_id,
        filters=filters,
    )
    if not ret["result"]:
        if "InvalidLaunchTemplateId.NotFound" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.launch_template", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    if not ret["ret"]["LaunchTemplates"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.launch_template", name=name
            )
        )
        return result

    resource = ret["ret"]["LaunchTemplates"][0]
    if len(ret["ret"]["LaunchTemplates"]) > 1:
        result["comment"].append(
            f"More than one aws.ec2.launch_template resource was found. Use resource {resource.get('LaunchTemplateId')}"
        )
    ret_version = (
        await hub.tool.aws.ec2.launch_template.search_raw_launch_template_version(
            ctx=ctx,
            resource_id=resource["LaunchTemplateId"],
        )
    )

    if not ret_version["result"]:
        result["result"] = False
        result["comment"] += list(ret["comment"])
        return result

    result[
        "ret"
    ] = hub.tool.aws.ec2.conversion_utils.convert_raw_lauchtemplate_to_present(
        raw_resource=resource,
        raw_version=ret_version["ret"]["LaunchTemplateVersions"][0],
        idem_resource_name=name,
    )

    return result


async def list_(
    hub,
    ctx,
    name: str = None,
    filters: List = None,
) -> Dict:
    """Get Launch templates from AWS. Supply one of the inputs as the filter.

    Args:
        name(str, Optional):
            The name of the Idem state.

        filters(list, Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_launch_templates

    Returns:
        Dict[str, Any]:
            Returns launch_templates in present format

    Examples:
        Calling this exec module function from the cli with filters

        .. code-block:: bash

            idem exec aws.ec2.launch_template.list name="my_resources" filters=[{'name': 'launch-template-name', 'values': ['template-name']}]

        Calling this exec module function from within a state module in pure python.

        .. code-block:: python

            async def state_function(hub, ctx, name, service_namespace, scaling_resource_id, scalable_dimension):
                await hub.exec.aws.ec2.launch_template.list(
                    ctx=ctx,
                    name=name,
                    filters=filters
                )

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ec2.launch_template.list
                - kwargs:
                    name: my_resources
                    filters:
                        - name: 'launch-template-name'
                          values: ["template-name"]
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.ec2.launch_template.search_raw(
        ctx=ctx,
        name=name,
        filters=filters,
    )
    if not ret["result"]:
        if "InvalidLaunchTemplateId.NotFound" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.launch_template", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    if not ret["ret"]["LaunchTemplates"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.launch_template", name=name
            )
        )
        return result
    result["ret"] = []
    for launch_template in ret["ret"]["LaunchTemplates"]:
        launch_template_id = launch_template.get("LaunchTemplateId")

        ret_version = (
            await hub.tool.aws.ec2.launch_template.search_raw_launch_template_version(
                ctx=ctx,
                resource_id=launch_template_id,
            )
        )
        if not ret_version["result"]:
            continue

        if not ret_version["ret"]["LaunchTemplateVersions"]:
            continue

        result_ret = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_lauchtemplate_to_present(
                raw_resource=launch_template,
                raw_version=ret_version["ret"]["LaunchTemplateVersions"][0],
                idem_resource_name=launch_template.get("LaunchTemplateName"),
            )
        )
        result["ret"].append(result_ret)
    return result



from typing import Dict


async def update(
    hub,
    ctx,
    resource_id,
    old_tags: Dict[str, str],
    new_tags: Dict[str, str],
):
    """
    Update tags of AWS Elasticsearch domain. No update is performed if new_tags is none

    Args:
        resource_id(str): AWS Elasticsearch domain resource id
        old_tags(Dict[str, str]): dict of old tags
        new_tags(Dict[str, str]): dict of new tags

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": dict of updated tags}

    """
    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )

    result = dict(comment=(), result=True, ret={})
    if (not tags_to_remove) and (not tags_to_add):
        return result

    account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    account_id = account_details["ret"]["Account"]
    resource_arn = hub.tool.aws.arn_utils.build(
        service="es",
        region=ctx["acct"]["region_name"],
        account_id=account_id,
        resource="domain/" + resource_id,
    )
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.es.remove_tags(
                ctx,
                ARN=resource_arn,
                TagKeys=[key for key, value in tags_to_remove.items()],
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.es.add_tags(
                ctx,
                ARN=resource_arn,
                TagList=hub.tool.aws.tag_utils.convert_tag_dict_to_list(
                    tags=tags_to_add
                ),
            )
            if not add_ret["result"]:
                result["comment"] += add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result



"""Util functions for AWS API Gateway v2 Route resources."""
from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_route_to_present(
    hub, api_id: str, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    """
    Convert AWS API Gateway v2 Route resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        api_id(str): The API resource identifier in Amazon Web Services.
        raw_resource(Dict[str, Any]): The AWS response to convert.
        idem_resource_name(str, Optional): An Idem name of the resource.

    Returns:
        Dict[str, Any]: Common idem present state
    """

    resource_parameters = OrderedDict(
        {
            "ApiGatewayManaged": "api_gateway_managed",
            "ApiKeyRequired": "api_key_required",
            "AuthorizationScopes": "authorization_scopes",
            "AuthorizationType": "authorization_type",
            "AuthorizerId": "authorizer_id",
            "ModelSelectionExpression": "model_selection_expression",
            "OperationName": "operation_name",
            "RequestModels": "request_models",
            "RequestParameters": "request_parameters",
            "RouteId": "route_id",
            "RouteKey": "route_key",
            "RouteResponseSelectionExpression": "route_response_selection_expression",
            "Target": "target",
        }
    )
    resource_translated = {
        "name": idem_resource_name
        if idem_resource_name
        else raw_resource.get("RouteId"),
        "resource_id": raw_resource.get("RouteId"),
        "api_id": api_id,
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def update(
    hub,
    ctx,
    api_id: str,
    resource_id: str,
    raw_resource: Dict[str, Any],
    resource_parameters: Dict[str, Any],
) -> Dict[str, Any]:
    """Updates a Route.

    Args:
        api_id(str):
            The API identifier.

        resource_id(str):
            The route ID.

        raw_resource(dict[str, Any]):
            Existing resource parameters in AWS.

        resource_parameters(dict[str, Any]):
            Parameters from SLS file.

    Returns:
        Dict[str, Any]:
            Returns the updated route.
    """
    result = dict(comment=(), result=True, ret=None)

    parameters = OrderedDict(
        {
            "ApiKeyRequired": "api_key_required",
            "AuthorizationScopes": "authorization_scopes",
            "AuthorizationType": "authorization_type",
            "AuthorizerId": "authorizer_id",
            "ModelSelectionExpression": "model_selection_expression",
            "OperationName": "operation_name",
            "RequestModels": "request_models",
            "RequestParameters": "request_parameters",
            "RouteKey": "route_key",
            "RouteResponseSelectionExpression": "route_response_selection_expression",
            "Target": "target",
        }
    )

    parameters_to_update = {}

    authorization_scopes = resource_parameters.get("AuthorizationScopes")
    if authorization_scopes is not None:
        if not hub.tool.aws.state_comparison_utils.are_lists_identical(
            authorization_scopes,
            raw_resource.get("AuthorizationScopes"),
        ):
            parameters_to_update["AuthorizationScopes"] = authorization_scopes

        resource_parameters.pop("AuthorizationScopes")

    for key, value in resource_parameters.items():
        if value is not None and value != raw_resource.get(key):
            parameters_to_update[key] = resource_parameters[key]

    if parameters_to_update:
        result["ret"] = {}
        for parameter_raw, parameter_present in parameters.items():
            if parameter_raw in parameters_to_update:
                result["ret"][parameter_present] = parameters_to_update[parameter_raw]

        if ctx.get("test", False):
            result["comment"] = (
                f"Would update parameters: " + ",".join(result["ret"].keys()),
            )
        else:
            update_ret = await hub.exec.boto3.client.apigatewayv2.update_route(
                ctx,
                ApiId=api_id,
                RouteId=resource_id,
                **parameters_to_update,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = update_ret["comment"]
                return result

            result["comment"] = (
                f"Updated parameters: " + ",".join(result["ret"].keys()),
            )

    return result
